import bluesky.plan_stubs as bps

from ophyd_async.core.device import Device
from ophyd_async.core.utils import DEFAULT_TIMEOUT, wait_for_connection


def ensure_connected(
    *devices: Device,
    mock: bool = False,
    timeout: float = DEFAULT_TIMEOUT,
    force_reconnect=False,
):
    yield from bps.wait_for(
        [
            lambda: wait_for_connection(
                **{
                    device.name: device.connect(mock, timeout, force_reconnect)
                    for device in devices
                }
            )
        ]
    )



import time
from typing import AsyncGenerator, AsyncIterator, Dict, Optional, Sequence
from unittest.mock import Mock

import bluesky.plan_stubs as bps
import pytest
from bluesky.protocols import DataKey, StreamAsset
from bluesky.run_engine import RunEngine
from event_model import ComposeStreamResourceBundle, compose_stream_resource

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    DetectorControl,
    DetectorWriter,
    HardwareTriggeredFlyable,
    observe_value,
    set_mock_value,
)
from ophyd_async.core.async_status import AsyncStatus, WatchableAsyncStatus
from ophyd_async.core.detector import StandardDetector
from ophyd_async.core.device import DeviceCollector
from ophyd_async.core.flyer import TriggerLogic
from ophyd_async.core.signal import SignalR
from ophyd_async.core.utils import WatcherUpdate
from ophyd_async.epics.pvi.pvi import fill_pvi_entries
from ophyd_async.epics.signal.signal import epics_signal_rw
from ophyd_async.panda import CommonPandaBlocks, StaticSeqTableTriggerLogic
from ophyd_async.plan_stubs import (
    prepare_static_seq_table_flyer_and_detectors_with_same_trigger,
    time_resolved_fly_and_collect_with_static_seq_table,
)
from ophyd_async.protocols import AsyncReadable


class DummyWriter(DetectorWriter):
    def __init__(self, name: str, shape: Sequence[int]):
        self.dummy_signal = epics_signal_rw(int, "pva://read_pv")
        self._shape = shape
        self._name = name
        self._file: Optional[ComposeStreamResourceBundle] = None
        self._last_emitted = 0
        self.index = 0

    async def open(self, multiplier: int = 1) -> Dict[str, DataKey]:
        return {
            self._name: DataKey(
                source="soft://some-source",
                shape=self._shape,
                dtype="number",
                external="STREAM:",
            )
        }

    async def observe_indices_written(
        self, timeout=DEFAULT_TIMEOUT
    ) -> AsyncGenerator[int, None]:
        num_captured: int
        async for num_captured in observe_value(self.dummy_signal, timeout):
            yield num_captured

    async def get_indices_written(self) -> int:
        return self.index

    async def collect_stream_docs(
        self, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        if indices_written:
            if not self._file:
                self._file = compose_stream_resource(
                    spec="AD_HDF5_SWMR_SLICE",
                    root="/",
                    data_key=self._name,
                    resource_path="",
                    resource_kwargs={
                        "path": "",
                        "multiplier": 1,
                        "timestamps": "/entry/instrument/NDAttributes/NDArrayTimeStamp",
                    },
                )
                yield "stream_resource", self._file.stream_resource_doc

            if indices_written >= self._last_emitted:
                indices = {
                    "start": self._last_emitted,
                    "stop": indices_written,
                }
                self._last_emitted = indices_written
                self._last_flush = time.monotonic()
                yield "stream_datum", self._file.compose_stream_datum(indices)

    async def close(self) -> None:
        self._file = None

    def increment_index(self) -> None:
        self.index += 1


class MockDetector(StandardDetector):
    def __init__(
        self,
        controller: DetectorControl,
        writer: DetectorWriter,
        config_sigs: Sequence[AsyncReadable] = [],
        name: str = "",
        writer_timeout: float = 1,
    ) -> None:
        super().__init__(controller, writer, config_sigs, name, writer_timeout)

    @WatchableAsyncStatus.wrap
    async def complete(self):
        assert self._arm_status, "Prepare not run"
        assert self._trigger_info
        self.writer.increment_index()
        async for index in self.writer.observe_indices_written(
            self._frame_writing_timeout
        ):
            yield WatcherUpdate(
                name=self.name,
                current=index,
                initial=self._initial_frame,
                target=self._trigger_info.num,
                unit="",
                precision=0,
                time_elapsed=time.monotonic() - self._fly_start,
            )
            if index >= self._trigger_info.num:
                break


@pytest.fixture
async def detectors(RE: RunEngine) -> tuple[MockDetector, MockDetector]:
    writers = [DummyWriter("testa", (1, 1)), DummyWriter("testb", (1, 1))]
    await writers[0].dummy_signal.connect(mock=True)
    await writers[1].dummy_signal.connect(mock=True)

    async def dummy_arm_1(self=None, trigger=None, num=0, exposure=None):
        return writers[0].dummy_signal.set(1)

    async def dummy_arm_2(self=None, trigger=None, num=0, exposure=None):
        return writers[1].dummy_signal.set(1)

    detector_1 = MockDetector(
        Mock(spec=DetectorControl, get_deadtime=lambda num: num, arm=dummy_arm_1),
        writers[0],
        name="detector_1",
        writer_timeout=3,
    )
    detector_2 = MockDetector(
        Mock(spec=DetectorControl, get_deadtime=lambda num: num, arm=dummy_arm_2),
        writers[1],
        name="detector_2",
        writer_timeout=3,
    )
    return (detector_1, detector_2)


@pytest.fixture
async def panda():
    class Panda(CommonPandaBlocks):
        def __init__(self, prefix: str, name: str = ""):
            self._prefix = prefix
            super().__init__(name)

        async def connect(self, mock: bool = False, timeout: float = DEFAULT_TIMEOUT):
            await fill_pvi_entries(
                self, self._prefix + "PVI", timeout=timeout, mock=mock
            )
            await super().connect(mock, timeout)

    async with DeviceCollector(mock=True):
        mock_panda = Panda("PANDAQSRV:", "mock_panda")

    assert mock_panda.name == "mock_panda"
    yield mock_panda


@pytest.fixture
async def flyer(panda):
    class MockFlyer(HardwareTriggeredFlyable):
        def __init__(
            self,
            trigger_logic: TriggerLogic,
            configuration_signals: Sequence[SignalR] = ...,
            name: str = "",
        ):
            super().__init__(trigger_logic, configuration_signals, name)

        @AsyncStatus.wrap
        async def kickoff(self) -> None:
            set_mock_value(self.trigger_logic.seq.active, 1)
            await super().kickoff()

        @AsyncStatus.wrap
        async def complete(self) -> None:
            set_mock_value(self.trigger_logic.seq.active, 0)
            await self._trigger_logic.complete()

    # Make flyer
    trigger_logic = StaticSeqTableTriggerLogic(panda.seq[1])
    flyer = MockFlyer(trigger_logic, [], name="flyer")

    return flyer


async def test_hardware_triggered_flyable_with_static_seq_table_logic(
    RE: RunEngine,
    detectors: tuple[StandardDetector],
    panda,
):
    """Run a dummy scan using a flyer with a prepare plan stub.

    This runs a dummy plan with two detectors and a flyer that uses
    StaticSeqTableTriggerLogic. The flyer and detectors are prepared with the
    prepare_static_seq_table_flyer_and_detectors_with_same_trigger plan stub.
    This stub creates trigger_info and a sequence table from given parameters
    and prepares the fly and both detectors with the same trigger info.

    """
    names = []
    docs = []
    detector_list = list(detectors)

    def append_and_print(name, doc):
        names.append(name)
        docs.append(doc)

    RE.subscribe(append_and_print)

    number_of_frames = 1
    exposure = 1
    shutter_time = 0.004

    trigger_logic = StaticSeqTableTriggerLogic(panda.seq[1])
    flyer = HardwareTriggeredFlyable(trigger_logic, [], name="flyer")

    def flying_plan():
        yield from bps.stage_all(*detector_list, flyer)

        yield from prepare_static_seq_table_flyer_and_detectors_with_same_trigger(
            flyer,
            detector_list,
            number_of_frames=number_of_frames,
            exposure=exposure,
            shutter_time=shutter_time,
        )

        for detector in detector_list:
            detector.controller.disarm.assert_called_once  # type: ignore

        yield from bps.open_run()
        yield from bps.declare_stream(*detector_list, name="main_stream", collect=True)

        set_mock_value(flyer.trigger_logic.seq.active, 1)

        yield from bps.kickoff(flyer, wait=True)
        for detector in detector_list:
            yield from bps.kickoff(detector)

        yield from bps.complete(flyer, wait=False, group="complete")
        for detector in detector_list:
            yield from bps.complete(detector, wait=False, group="complete")

        # Manually incremenet the index as if a frame was taken
        for detector in detector_list:
            detector.writer.increment_index()

        set_mock_value(flyer.trigger_logic.seq.active, 0)

        done = False
        while not done:
            try:
                yield from bps.wait(group="complete", timeout=0.5)
            except TimeoutError:
                pass
            else:
                done = True
            yield from bps.collect(
                *detector_list,
                return_payload=False,
                name="main_stream",
            )
        yield from bps.wait(group="complete")
        yield from bps.close_run()

        yield from bps.unstage_all(flyer, *detector_list)
        for detector in detector_list:
            assert detector.controller.disarm.called  # type: ignore

    # fly scan
    RE(flying_plan())

    assert names == [
        "start",
        "descriptor",
        "stream_resource",
        "stream_datum",
        "stream_resource",
        "stream_datum",
        "stop",
    ]


async def test_time_resolved_fly_and_collect_with_static_seq_table(
    RE: RunEngine,
    detectors: tuple[StandardDetector],
    flyer,
):
    names = []
    docs = []
    detector_list = list(detectors)

    def append_and_print(name, doc):
        names.append(name)
        docs.append(doc)

    RE.subscribe(append_and_print)

    # Trigger parameters
    number_of_frames = 1
    exposure = 1
    shutter_time = 0.004

    def fly():
        yield from bps.stage_all(*detector_list, flyer)
        yield from bps.open_run()
        yield from time_resolved_fly_and_collect_with_static_seq_table(
            stream_name="stream1",
            flyer=flyer,
            detectors=detector_list,
            number_of_frames=number_of_frames,
            exposure=exposure,
            shutter_time=shutter_time,
        )
        yield from bps.close_run()
        yield from bps.unstage_all(flyer, *detector_list)

    # fly scan
    RE(fly())

    assert names == [
        "start",
        "descriptor",
        "stream_resource",
        "stream_datum",
        "stream_resource",
        "stream_datum",
        "stop",
    ]


@pytest.mark.parametrize("detector_list", [[], None])
async def test_at_least_one_detector_in_fly_plan(
    RE: RunEngine,
    flyer,
    detector_list,
):
    # Trigger parameters
    number_of_frames = 1
    exposure = 1
    shutter_time = 0.004

    assert not detector_list

    def fly():
        yield from time_resolved_fly_and_collect_with_static_seq_table(
            stream_name="stream1",
            flyer=flyer,
            detectors=detector_list,
            number_of_frames=number_of_frames,
            exposure=exposure,
            shutter_time=shutter_time,
        )

    with pytest.raises(ValueError) as exc:
        RE(fly())
        assert str(exc) == "No detectors provided. There must be at least one."



import asyncio
from typing import Optional, Set

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    DetectorControl,
    DetectorTrigger,
)

from ..drivers.ad_base import (
    DEFAULT_GOOD_STATES,
    ADBase,
    DetectorState,
    ImageMode,
    start_acquiring_driver_and_ensure_status,
)
from ..utils import stop_busy_record


class ADSimController(DetectorControl):
    def __init__(
        self, driver: ADBase, good_states: Set[DetectorState] = set(DEFAULT_GOOD_STATES)
    ) -> None:
        self.driver = driver
        self.good_states = good_states

    def get_deadtime(self, exposure: float) -> float:
        return 0.002

    async def arm(
        self,
        num: int,
        trigger: DetectorTrigger = DetectorTrigger.internal,
        exposure: Optional[float] = None,
    ) -> AsyncStatus:
        assert (
            trigger == DetectorTrigger.internal
        ), "fly scanning (i.e. external triggering) is not supported for this device"
        frame_timeout = DEFAULT_TIMEOUT + await self.driver.acquire_time.get_value()
        await asyncio.gather(
            self.driver.num_images.set(num),
            self.driver.image_mode.set(ImageMode.multiple),
        )
        return await start_acquiring_driver_and_ensure_status(
            self.driver, good_states=self.good_states, timeout=frame_timeout
        )

    async def disarm(self):
        # We can't use caput callback as we already used it in arm() and we can't have
        # 2 or they will deadlock
        await stop_busy_record(self.driver.acquire, False, timeout=1)



from enum import Enum
from typing import Optional
from xml.etree import cElementTree as ET

from ophyd_async.core import DEFAULT_TIMEOUT, SignalRW, T, wait_for_value


class FileWriteMode(str, Enum):
    single = "Single"
    capture = "Capture"
    stream = "Stream"


class ImageMode(str, Enum):
    single = "Single"
    multiple = "Multiple"
    continuous = "Continuous"


class NDAttributeDataType(str, Enum):
    INT = "INT"
    DOUBLE = "DOUBLE"
    STRING = "STRING"


class NDAttributesXML:
    """Helper to make NDAttributesFile XML for areaDetector"""

    _dbr_types = {
        None: "DBR_NATIVE",
        NDAttributeDataType.INT: "DBR_LONG",
        NDAttributeDataType.DOUBLE: "DBR_DOUBLE",
        NDAttributeDataType.STRING: "DBR_STRING",
    }

    def __init__(self):
        self._root = ET.Element("Attributes")

    def add_epics_pv(
        self,
        name: str,
        pv: str,
        datatype: Optional[NDAttributeDataType] = None,
        description: str = "",
    ):
        """Add a PV to the attribute list

        Args:
            name: The attribute name
            pv: The pv to get from
            datatype: An override datatype, otherwise will use native EPICS type
            description: A description that appears in the HDF file as an attribute
        """
        ET.SubElement(
            self._root,
            "Attribute",
            name=name,
            type="EPICS_PV",
            source=pv,
            datatype=self._dbr_types[datatype],
            description=description,
        )

    def add_param(
        self,
        name: str,
        param: str,
        datatype: NDAttributeDataType,
        addr: int = 0,
        description: str = "",
    ):
        """Add a driver or plugin parameter to the attribute list

        Args:
            name: The attribute name
            param: The parameter string as seen in the INP link of the record
            datatype: The datatype of the parameter
            description: A description that appears in the HDF file as an attribute
        """
        ET.SubElement(
            self._root,
            "Attribute",
            name=name,
            type="PARAM",
            source=param,
            addr=str(addr),
            datatype=datatype.value,
            description=description,
        )

    def __str__(self) -> str:
        """Output the XML pretty printed"""
        ET.indent(self._root, space="    ", level=0)
        return ET.tostring(self._root, xml_declaration=True, encoding="utf-8").decode()


async def stop_busy_record(
    signal: SignalRW[T],
    value: T,
    timeout: float = DEFAULT_TIMEOUT,
    status_timeout: Optional[float] = None,
) -> None:
    await signal.set(value, wait=False, timeout=status_timeout)
    await wait_for_value(signal, value, timeout=timeout)



from enum import Enum

import pytest

from ophyd_async.epics._backend.common import get_supported_values


def test_given_a_non_enum_passed_to_get_supported_enum_then_raises():
    with pytest.raises(TypeError):
        get_supported_values("", int, ("test",))


def test_given_an_enum_but_not_str_passed_to_get_supported_enum_then_raises():
    class MyEnum(Enum):
        TEST = "test"

    with pytest.raises(TypeError):
        get_supported_values("", MyEnum, ("test",))


def test_given_pv_has_choices_not_in_supplied_enum_then_raises():
    class MyEnum(str, Enum):
        TEST = "test"

    with pytest.raises(TypeError):
        get_supported_values("", MyEnum, ("test", "unexpected_choice"))


def test_given_supplied_enum_has_choices_not_in_pv_then_raises():
    class MyEnum(str, Enum):
        TEST = "test"
        OTHER = "unexpected_choice"

    with pytest.raises(TypeError):
        get_supported_values("", MyEnum, ("test",))


def test_given_no_supplied_enum_then_returns_generated_choices_enum_with_pv_choices():
    supported_vals = get_supported_values("", None, ("test",))
    assert len(supported_vals) == 1
    assert "test" in supported_vals


def test_given_a_supplied_enum_that_matches_the_pv_choices_then_enum_type_is_returned():
    class MyEnum(str, Enum):
        TEST_1 = "test_1"
        TEST_2 = "test_2"

    supported_vals = get_supported_values("", MyEnum, ("test_1", "test_2"))
    assert len(supported_vals) == 2
    assert "test_1" in supported_vals
    assert "test_2" in supported_vals



from typing import AsyncGenerator, AsyncIterator, Dict

from bluesky.protocols import DataKey

from ophyd_async.core import DirectoryProvider
from ophyd_async.core.detector import DetectorWriter
from ophyd_async.sim.pattern_generator import PatternGenerator


class SimPatternDetectorWriter(DetectorWriter):
    pattern_generator: PatternGenerator

    def __init__(
        self, pattern_generator: PatternGenerator, directoryProvider: DirectoryProvider
    ) -> None:
        self.pattern_generator = pattern_generator
        self.directory_provider = directoryProvider

    async def open(self, multiplier: int = 1) -> Dict[str, DataKey]:
        return await self.pattern_generator.open_file(
            self.directory_provider, multiplier
        )

    async def close(self) -> None:
        self.pattern_generator.close()

    def collect_stream_docs(self, indices_written: int) -> AsyncIterator:
        return self.pattern_generator.collect_stream_docs(indices_written)

    def observe_indices_written(self, timeout=...) -> AsyncGenerator[int, None]:
        return self.pattern_generator.observe_indices_written()

    async def get_indices_written(self) -> int:
        return self.pattern_generator.written_images_counter



"""Test file specifying how we want to interact with the panda controller"""

from unittest.mock import patch

import pytest

from ophyd_async.core import DEFAULT_TIMEOUT, DetectorTrigger, Device, DeviceCollector
from ophyd_async.epics.pvi import fill_pvi_entries
from ophyd_async.epics.signal import epics_signal_rw
from ophyd_async.panda import CommonPandaBlocks, PandaPcapController


@pytest.fixture
async def mock_panda():
    class Panda(CommonPandaBlocks):
        def __init__(self, prefix: str, name: str = ""):
            self._prefix = prefix
            super().__init__(name)

        async def connect(self, mock: bool = False, timeout: float = DEFAULT_TIMEOUT):
            await fill_pvi_entries(
                self, self._prefix + "PVI", timeout=timeout, mock=mock
            )
            await super().connect(mock=mock, timeout=timeout)

    async with DeviceCollector(mock=True):
        mock_panda = Panda("PANDACONTROLLER:", name="mock_panda")
        mock_panda.phase_1_signal_units = epics_signal_rw(int, "")
    yield mock_panda


async def test_panda_controller_not_filled_blocks():
    class PcapBlock(Device):
        pass  # Not filled

    pandaController = PandaPcapController(pcap=PcapBlock())
    with patch("ophyd_async.panda._panda_controller.wait_for_value", return_value=None):
        with pytest.raises(AttributeError) as exc:
            await pandaController.arm(num=1, trigger=DetectorTrigger.constant_gate)
    assert ("'PcapBlock' object has no attribute 'arm'") in str(exc.value)


async def test_panda_controller_arm_disarm(mock_panda):
    pandaController = PandaPcapController(mock_panda.pcap)
    with patch("ophyd_async.panda._panda_controller.wait_for_value", return_value=None):
        await pandaController.arm(num=1, trigger=DetectorTrigger.constant_gate)
    await pandaController.disarm()


async def test_panda_controller_wrong_trigger():
    pandaController = PandaPcapController(None)
    with pytest.raises(AssertionError):
        await pandaController.arm(num=1, trigger=DetectorTrigger.internal)



"""Module which defines abstract classes to work with detectors"""

import asyncio
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import (
    AsyncGenerator,
    AsyncIterator,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    TypeVar,
)

from bluesky.protocols import (
    Collectable,
    DataKey,
    Flyable,
    Preparable,
    Reading,
    Stageable,
    StreamAsset,
    Triggerable,
    WritesStreamAssets,
)

from ophyd_async.protocols import AsyncConfigurable, AsyncReadable

from .async_status import AsyncStatus, WatchableAsyncStatus
from .device import Device
from .utils import DEFAULT_TIMEOUT, WatcherUpdate, merge_gathered_dicts

T = TypeVar("T")


class DetectorTrigger(str, Enum):
    """Type of mechanism for triggering a detector to take frames"""

    #: Detector generates internal trigger for given rate
    internal = "internal"
    #: Expect a series of arbitrary length trigger signals
    edge_trigger = "edge_trigger"
    #: Expect a series of constant width external gate signals
    constant_gate = "constant_gate"
    #: Expect a series of variable width external gate signals
    variable_gate = "variable_gate"


@dataclass(frozen=True)
class TriggerInfo:
    """Minimal set of information required to setup triggering on a detector"""

    #: Number of triggers that will be sent
    num: int
    #: Sort of triggers that will be sent
    trigger: DetectorTrigger
    #: What is the minimum deadtime between triggers
    deadtime: float
    #: What is the maximum high time of the triggers
    livetime: float


class DetectorControl(ABC):
    """
    Classes implementing this interface should hold the logic for
    arming and disarming a detector
    """

    @abstractmethod
    def get_deadtime(self, exposure: float) -> float:
        """For a given exposure, how long should the time between exposures be"""

    @abstractmethod
    async def arm(
        self,
        num: int,
        trigger: DetectorTrigger = DetectorTrigger.internal,
        exposure: Optional[float] = None,
    ) -> AsyncStatus:
        """
        Arm detector, do all necessary steps to prepare detector for triggers.

        Args:
            num: Expected number of frames
            trigger: Type of trigger for which to prepare the detector. Defaults to
            DetectorTrigger.internal.
            exposure: Exposure time with which to set up the detector. Defaults to None
            if not applicable or the detector is expected to use its previously-set
            exposure time.

        Returns:
            AsyncStatus: Status representing the arm operation. This function returning
            represents the start of the arm. The returned status completing means
            the detector is now armed.
        """

    @abstractmethod
    async def disarm(self):
        """Disarm the detector, return detector to an idle state"""


class DetectorWriter(ABC):
    """Logic for making a detector write data to somewhere persistent
    (e.g. an HDF5 file)"""

    @abstractmethod
    async def open(self, multiplier: int = 1) -> Dict[str, DataKey]:
        """Open writer and wait for it to be ready for data.

        Args:
            multiplier: Each StreamDatum index corresponds to this many
                written exposures

        Returns:
            Output for ``describe()``
        """

    @abstractmethod
    def observe_indices_written(
        self, timeout=DEFAULT_TIMEOUT
    ) -> AsyncGenerator[int, None]:
        """Yield the index of each frame (or equivalent data point) as it is written"""

    @abstractmethod
    async def get_indices_written(self) -> int:
        """Get the number of indices written"""

    @abstractmethod
    def collect_stream_docs(self, indices_written: int) -> AsyncIterator[StreamAsset]:
        """Create Stream docs up to given number written"""

    @abstractmethod
    async def close(self) -> None:
        """Close writer, blocks until I/O is complete"""


class StandardDetector(
    Device,
    Stageable,
    AsyncConfigurable,
    AsyncReadable,
    Triggerable,
    Preparable,
    Flyable,
    Collectable,
    WritesStreamAssets,
    Generic[T],
):
    """
    Useful detector base class for step and fly scanning detectors.
    Aggregates controller and writer logic together.
    """

    def __init__(
        self,
        controller: DetectorControl,
        writer: DetectorWriter,
        config_sigs: Sequence[AsyncReadable] = (),
        name: str = "",
        writer_timeout: float = DEFAULT_TIMEOUT,
    ) -> None:
        """
        Constructor

        Args:
            controller: Logic for arming and disarming the detector
            writer: Logic for making the detector write persistent data
            config_sigs: Signals to read when describe and read
            configuration are called. Defaults to ().
            name: Device name. Defaults to "".
            writer_timeout: Timeout for frame writing to start, if the
            timeout is reached, ophyd-async assumes the detector
            has a problem and raises an error.
            Defaults to DEFAULT_TIMEOUT.
        """
        self._controller = controller
        self._writer = writer
        self._describe: Dict[str, DataKey] = {}
        self._config_sigs = list(config_sigs)
        self._frame_writing_timeout = writer_timeout
        # For prepare
        self._arm_status: Optional[AsyncStatus] = None
        self._trigger_info: Optional[TriggerInfo] = None
        # For kickoff
        self._watchers: List[Callable] = []
        self._fly_status: Optional[WatchableAsyncStatus] = None
        self._fly_start: float

        self._intial_frame: int
        self._last_frame: int
        super().__init__(name)

    @property
    def controller(self) -> DetectorControl:
        return self._controller

    @property
    def writer(self) -> DetectorWriter:
        return self._writer

    @AsyncStatus.wrap
    async def stage(self) -> None:
        # Disarm the detector, stop filewriting, and open file for writing.
        await self._check_config_sigs()
        await asyncio.gather(self.writer.close(), self.controller.disarm())
        self._describe = await self.writer.open()

    async def _check_config_sigs(self):
        """Checks configuration signals are named and connected."""
        for signal in self._config_sigs:
            if signal.name == "":
                raise Exception(
                    "config signal must be named before it is passed to the detector"
                )
            try:
                await signal.get_value()
            except NotImplementedError:
                raise Exception(
                    f"config signal {signal._name} must be connected before it is "
                    + "passed to the detector"
                )

    @AsyncStatus.wrap
    async def unstage(self) -> None:
        # Stop data writing.
        await self.writer.close()

    async def read_configuration(self) -> Dict[str, Reading]:
        return await merge_gathered_dicts(sig.read() for sig in self._config_sigs)

    async def describe_configuration(self) -> Dict[str, DataKey]:
        return await merge_gathered_dicts(sig.describe() for sig in self._config_sigs)

    async def read(self) -> Dict[str, Reading]:
        # All data is in StreamResources, not Events, so nothing to output here
        return {}

    async def describe(self) -> Dict[str, DataKey]:
        return self._describe

    @AsyncStatus.wrap
    async def trigger(self) -> None:
        # Arm the detector and wait for it to finish.
        indices_written = await self.writer.get_indices_written()
        written_status = await self.controller.arm(
            num=1,
            trigger=DetectorTrigger.internal,
        )
        await written_status
        end_observation = indices_written + 1

        async for index in self.writer.observe_indices_written(
            self._frame_writing_timeout
        ):
            if index >= end_observation:
                break

    def prepare(
        self,
        value: T,
    ) -> AsyncStatus:
        # Just arm detector for the time being
        return AsyncStatus(self._prepare(value))

    async def _prepare(self, value: T) -> None:
        """
        Arm detector.

        Prepare the detector with trigger information. This is determined at and passed
        in from the plan level.

        This currently only prepares detectors for flyscans and stepscans just use the
        trigger information determined in trigger.

        To do: Unify prepare to be use for both fly and step scans.

        Args:
            value: TriggerInfo describing how to trigger the detector
        """
        assert type(value) is TriggerInfo
        self._trigger_info = value
        self._initial_frame = await self.writer.get_indices_written()
        self._last_frame = self._initial_frame + self._trigger_info.num

        required = self.controller.get_deadtime(self._trigger_info.livetime)
        assert required <= self._trigger_info.deadtime, (
            f"Detector {self.controller} needs at least {required}s deadtime, "
            f"but trigger logic provides only {self._trigger_info.deadtime}s"
        )
        self._arm_status = await self.controller.arm(
            num=self._trigger_info.num,
            trigger=self._trigger_info.trigger,
            exposure=self._trigger_info.livetime,
        )
        self._fly_start = time.monotonic()

    @AsyncStatus.wrap
    async def kickoff(self):
        if not self._arm_status:
            raise Exception("Detector not armed!")

    @WatchableAsyncStatus.wrap
    async def complete(self):
        assert self._arm_status, "Prepare not run"
        assert self._trigger_info
        async for index in self.writer.observe_indices_written(
            self._frame_writing_timeout
        ):
            yield WatcherUpdate(
                name=self.name,
                current=index,
                initial=self._initial_frame,
                target=self._trigger_info.num,
                unit="",
                precision=0,
                time_elapsed=time.monotonic() - self._fly_start,
            )
            if index >= self._trigger_info.num:
                break

    async def describe_collect(self) -> Dict[str, DataKey]:
        return self._describe

    async def collect_asset_docs(
        self, index: Optional[int] = None
    ) -> AsyncIterator[StreamAsset]:
        # Collect stream datum documents for all indices written.
        # The index is optional, and provided for fly scans, however this needs to be
        # retrieved for step scans.
        if index is None:
            index = await self.writer.get_indices_written()
        async for doc in self.writer.collect_stream_docs(index):
            yield doc

    async def get_index(self) -> int:
        return await self.writer.get_indices_written()



from enum import Enum

from ophyd_async.epics.signal.signal import epics_signal_rw_rbv

from .ad_base import ADBase


class PilatusTriggerMode(str, Enum):
    internal = "Internal"
    ext_enable = "Ext. Enable"
    ext_trigger = "Ext. Trigger"
    mult_trigger = "Mult. Trigger"
    alignment = "Alignment"


class PilatusDriver(ADBase):
    def __init__(self, prefix: str, name: str = "") -> None:
        self.trigger_mode = epics_signal_rw_rbv(
            PilatusTriggerMode, prefix + "TriggerMode"
        )
        super().__init__(prefix, name)



from ._common_blocks import (
    CommonPandaBlocks,
    DataBlock,
    PcapBlock,
    PulseBlock,
    SeqBlock,
    TimeUnits,
)
from ._hdf_panda import HDFPanda
from ._panda_controller import PandaPcapController
from ._table import (
    SeqTable,
    SeqTableRow,
    SeqTrigger,
    seq_table_from_arrays,
    seq_table_from_rows,
)
from ._trigger import StaticSeqTableTriggerLogic
from ._utils import phase_sorter

__all__ = [
    "CommonPandaBlocks",
    "HDFPanda",
    "PcapBlock",
    "PulseBlock",
    "seq_table_from_arrays",
    "seq_table_from_rows",
    "SeqBlock",
    "SeqTable",
    "SeqTableRow",
    "SeqTrigger",
    "phase_sorter",
    "PandaPcapController",
    "TimeUnits",
    "DataBlock",
    "CommonPandABlocks",
    "StaticSeqTableTriggerLogic",
]



from typing import Sequence
from unittest.mock import patch

import pytest

from ophyd_async.core import (
    DeviceCollector,
    ShapeProvider,
    StaticDirectoryProvider,
    set_mock_value,
)
from ophyd_async.epics.areadetector.writers import HDFWriter, NDFileHDF


class DummyShapeProvider(ShapeProvider):
    def __init__(self) -> None:
        pass

    async def __call__(self) -> Sequence[int]:
        return (10, 10)


@pytest.fixture
async def hdf_writer(RE) -> HDFWriter:
    async with DeviceCollector(mock=True):
        hdf = NDFileHDF("HDF:")

    return HDFWriter(
        hdf,
        StaticDirectoryProvider("some_path", "some_prefix"),
        name_provider=lambda: "test",
        shape_provider=DummyShapeProvider(),
    )


async def test_correct_descriptor_doc_after_open(hdf_writer: HDFWriter):
    set_mock_value(hdf_writer.hdf.file_path_exists, True)
    with patch("ophyd_async.core.signal.wait_for_value", return_value=None):
        descriptor = await hdf_writer.open()

    assert descriptor == {
        "test": {
            "source": "mock+ca://HDF:FullFileName_RBV",
            "shape": (10, 10),
            "dtype": "array",
            "external": "STREAM:",
        }
    }

    await hdf_writer.close()


async def test_collect_stream_docs(hdf_writer: HDFWriter):
    assert hdf_writer._file is None

    [item async for item in hdf_writer.collect_stream_docs(1)]
    assert hdf_writer._file



from enum import Enum
from functools import partial
from typing import Any, Callable, Dict, Generator, List, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt
import yaml
from bluesky.plan_stubs import abs_set, wait
from bluesky.protocols import Location
from bluesky.utils import Msg
from epicscorelibs.ca.dbr import ca_array, ca_float, ca_int, ca_str

from .device import Device
from .signal import SignalRW

CaType = Union[ca_float, ca_int, ca_str, ca_array]


def ndarray_representer(dumper: yaml.Dumper, array: npt.NDArray[Any]) -> yaml.Node:
    return dumper.represent_sequence(
        "tag:yaml.org,2002:seq", array.tolist(), flow_style=True
    )


def ca_dbr_representer(dumper: yaml.Dumper, value: CaType) -> yaml.Node:
    # if it's an array, just call ndarray_representer...
    represent_array = partial(ndarray_representer, dumper)

    representers: Dict[CaType, Callable[[CaType], yaml.Node]] = {
        ca_float: dumper.represent_float,
        ca_int: dumper.represent_int,
        ca_str: dumper.represent_str,
        ca_array: represent_array,
    }
    return representers[type(value)](value)


class OphydDumper(yaml.Dumper):
    def represent_data(self, data: Any) -> Any:
        if isinstance(data, Enum):
            return self.represent_data(data.value)
        return super(OphydDumper, self).represent_data(data)


def get_signal_values(
    signals: Dict[str, SignalRW[Any]], ignore: Optional[List[str]] = None
) -> Generator[Msg, Sequence[Location[Any]], Dict[str, Any]]:
    """Get signal values in bulk.

    Used as part of saving the signals of a device to a yaml file.

    Parameters
    ----------
    signals : Dict[str, SignalRW]
        Dictionary with pv names and matching SignalRW values. Often the direct result
        of :func:`walk_rw_signals`.

    ignore : Optional[List[str]]
        Optional list of PVs that should be ignored.

    Returns
    -------
    Dict[str, Any]
        A dictionary containing pv names and their associated values. Ignored pvs are
        set to None.

    See Also
    --------
    :func:`ophyd_async.core.walk_rw_signals`
    :func:`ophyd_async.core.save_to_yaml`
    """

    ignore = ignore or []
    selected_signals = {
        key: signal for key, signal in signals.items() if key not in ignore
    }
    selected_values = yield Msg("locate", *selected_signals.values())

    # TODO: investigate wrong type hints
    if isinstance(selected_values, dict):
        selected_values = [selected_values]  # type: ignore

    assert selected_values is not None, "No signalRW's were able to be located"
    named_values = {
        key: value["setpoint"] for key, value in zip(selected_signals, selected_values)
    }
    # Ignored values place in with value None so we know which ones were ignored
    named_values.update({key: None for key in ignore})
    return named_values


def walk_rw_signals(
    device: Device, path_prefix: Optional[str] = ""
) -> Dict[str, SignalRW[Any]]:
    """Retrieve all SignalRWs from a device.

    Stores retrieved signals with their dotted attribute paths in a dictionary. Used as
    part of saving and loading a device.

    Parameters
    ----------
    device : Device
        Ophyd device to retrieve read-write signals from.

    path_prefix : str
        For internal use, leave blank when calling the method.

    Returns
    -------
    SignalRWs : dict
        A dictionary matching the string attribute path of a SignalRW with the
        signal itself.

        See Also
    --------
    :func:`ophyd_async.core.get_signal_values`
    :func:`ophyd_async.core.save_to_yaml`

    """

    if not path_prefix:
        path_prefix = ""

    signals: Dict[str, SignalRW[Any]] = {}
    for attr_name, attr in device.children():
        dot_path = f"{path_prefix}{attr_name}"
        if type(attr) is SignalRW:
            signals[dot_path] = attr
        attr_signals = walk_rw_signals(attr, path_prefix=dot_path + ".")
        signals.update(attr_signals)
    return signals


def save_to_yaml(phases: Sequence[Dict[str, Any]], save_path: str) -> None:
    """Plan which serialises a phase or set of phases of SignalRWs to a yaml file.

    Parameters
    ----------
    phases : dict or list of dicts
        The values to save. Each item in the list is a seperate phase used when loading
        a device. In general this variable be the return value of `get_signal_values`.

    save_path : str
        Path of the yaml file to write to

    See Also
    --------
    :func:`ophyd_async.core.walk_rw_signals`
    :func:`ophyd_async.core.get_signal_values`
    :func:`ophyd_async.core.load_from_yaml`
    """

    yaml.add_representer(np.ndarray, ndarray_representer, Dumper=yaml.Dumper)

    yaml.add_representer(ca_float, ca_dbr_representer, Dumper=yaml.Dumper)
    yaml.add_representer(ca_int, ca_dbr_representer, Dumper=yaml.Dumper)
    yaml.add_representer(ca_str, ca_dbr_representer, Dumper=yaml.Dumper)
    yaml.add_representer(ca_array, ca_dbr_representer, Dumper=yaml.Dumper)

    with open(save_path, "w") as file:
        yaml.dump(phases, file, Dumper=OphydDumper, default_flow_style=False)


def load_from_yaml(save_path: str) -> Sequence[Dict[str, Any]]:
    """Plan that returns a list of dicts with saved signal values from a yaml file.

    Parameters
    ----------
    save_path : str
        Path of the yaml file to load from

    See Also
    --------
    :func:`ophyd_async.core.save_to_yaml`
    :func:`ophyd_async.core.set_signal_values`
    """
    with open(save_path, "r") as file:
        return yaml.full_load(file)


def set_signal_values(
    signals: Dict[str, SignalRW[Any]], values: Sequence[Dict[str, Any]]
) -> Generator[Msg, None, None]:
    """Maps signals from a yaml file into device signals.

    ``values`` contains signal values in phases, which are loaded in sequentially
    into the provided signals, to ensure signals are set in the correct order.

    Parameters
    ----------
    signals : Dict[str, SignalRW[Any]]
        Dictionary of named signals to be updated if value found in values argument.
        Can be the output of :func:`walk_rw_signals()` for a device.

    values : Sequence[Dict[str, Any]]
        List of dictionaries of signal name and value pairs, if a signal matches
        the name of one in the signals argument, sets the signal to that value.
        The groups of signals are loaded in their list order.
        Can be the output of :func:`load_from_yaml()` for a yaml file.

    See Also
    --------
    :func:`ophyd_async.core.load_from_yaml`
    :func:`ophyd_async.core.walk_rw_signals`
    """
    # For each phase, set all the signals,
    # load them to the correct value and wait for the load to complete
    for phase_number, phase in enumerate(values):
        # Key is signal name
        for key, value in phase.items():
            # Skip ignored values
            if value is None:
                continue

            if key in signals:
                yield from abs_set(
                    signals[key], value, group=f"load-phase{phase_number}"
                )

        yield from wait(f"load-phase{phase_number}")


def load_device(device: Device, path: str):
    """Plan which loads PVs from a yaml file into a device.

    Parameters
    ----------
    device: Device
        The device to load PVs into
    path: str
        Path of the yaml file to load

    See Also
    --------
    :func:`ophyd_async.core.save_device`
    """
    values = load_from_yaml(path)
    signals_to_set = walk_rw_signals(device)
    yield from set_signal_values(signals_to_set, values)


def all_at_once(values: Dict[str, Any]) -> Sequence[Dict[str, Any]]:
    """Sort all the values into a single phase so they are set all at once"""
    return [values]


def save_device(
    device: Device,
    path: str,
    sorter: Callable[[Dict[str, Any]], Sequence[Dict[str, Any]]] = all_at_once,
    ignore: Optional[List[str]] = None,
):
    """Plan that saves the state of all PV's on a device using a sorter.

    The default sorter assumes all saved PVs can be loaded at once, and therefore
    can be saved at one time, i.e. all PVs will appear on one list in the
    resulting yaml file.

    This can be a problem, because when the yaml is ingested with
    :func:`ophyd_async.core.load_device`, it will set all of those PVs at once.
    However, some PV's need to be set before others - this is device specific.

    Therefore, users should consider the order of device loading and write their
    own sorter algorithms accordingly.

    See :func:`ophyd_async.panda.phase_sorter` for a valid implementation of the
    sorter.

    Parameters
    ----------
    device : Device
        The device whose PVs should be saved.

    path : str
        The path where the resulting yaml should be saved to

    sorter : Callable[[Dict[str, Any]], Sequence[Dict[str, Any]]]

    ignore : Optional[List[str]]

    See Also
    --------
    :func:`ophyd_async.core.load_device`
    """
    values = yield from get_signal_values(walk_rw_signals(device), ignore=ignore)
    save_to_yaml(sorter(values), path)



from enum import Enum

from ophyd_async.epics.signal.signal import epics_signal_rw_rbv

from .ad_base import ADBase


class VimbaPixelFormat(str, Enum):
    internal = "Mono8"
    ext_enable = "Mono12"
    ext_trigger = "Ext. Trigger"
    mult_trigger = "Mult. Trigger"
    alignment = "Alignment"


class VimbaConvertFormat(str, Enum):
    none = "None"
    mono8 = "Mono8"
    mono16 = "Mono16"
    rgb8 = "RGB8"
    rgb16 = "RGB16"


class VimbaTriggerSource(str, Enum):
    freerun = "Freerun"
    line1 = "Line1"
    line2 = "Line2"
    fixed_rate = "FixedRate"
    software = "Software"
    action0 = "Action0"
    action1 = "Action1"


class VimbaOverlap(str, Enum):
    off = "Off"
    prev_frame = "PreviousFrame"


class VimbaOnOff(str, Enum):
    on = "On"
    off = "Off"


class VimbaExposeOutMode(str, Enum):
    timed = "Timed"  # Use ExposureTime PV
    trigger_width = "TriggerWidth"  # Expose for length of high signal


class VimbaDriver(ADBase):
    def __init__(self, prefix: str, name: str = "") -> None:
        # self.pixel_format = epics_signal_rw_rbv(PixelFormat, prefix + "PixelFormat")
        self.convert_format = epics_signal_rw_rbv(
            VimbaConvertFormat, prefix + "ConvertPixelFormat"
        )  # Pixel format of data outputted to AD
        self.trig_source = epics_signal_rw_rbv(
            VimbaTriggerSource, prefix + "TriggerSource"
        )
        self.trigger_mode = epics_signal_rw_rbv(VimbaOnOff, prefix + "TriggerMode")
        self.overlap = epics_signal_rw_rbv(VimbaOverlap, prefix + "TriggerOverlap")
        self.expose_mode = epics_signal_rw_rbv(
            VimbaExposeOutMode, prefix + "ExposureMode"
        )
        super().__init__(prefix, name)



from typing import Any, Dict, Sequence


def phase_sorter(panda_signal_values: Dict[str, Any]) -> Sequence[Dict[str, Any]]:
    # Panda has two load phases. If the signal name ends in the string "UNITS",
    # it needs to be loaded first so put in first phase
    phase_1, phase_2 = {}, {}

    for key, value in panda_signal_values.items():
        if key.endswith("units"):
            phase_1[key] = value
        else:
            phase_2[key] = value

    return [phase_1, phase_2]



import pytest
from bluesky.run_engine import RunEngine

from ophyd_async.core import (
    DetectorTrigger,
    DeviceCollector,
    DirectoryProvider,
    set_mock_value,
)
from ophyd_async.epics.areadetector.kinetix import KinetixDetector


@pytest.fixture
async def adkinetix(
    RE: RunEngine,
    static_directory_provider: DirectoryProvider,
) -> KinetixDetector:
    async with DeviceCollector(mock=True):
        adkinetix = KinetixDetector("KINETIX:", static_directory_provider)

    return adkinetix


async def test_get_deadtime(
    adkinetix: KinetixDetector,
):
    # Currently Kinetix driver doesn't support getting deadtime.
    assert adkinetix._controller.get_deadtime(0) == 0.001


async def test_trigger_modes(adkinetix: KinetixDetector):
    set_mock_value(adkinetix.drv.trigger_mode, "Internal")

    async def setup_trigger_mode(trig_mode: DetectorTrigger):
        await adkinetix.controller.arm(num=1, trigger=trig_mode)
        # Prevent timeouts
        set_mock_value(adkinetix.drv.acquire, True)

    # Default TriggerSource
    assert (await adkinetix.drv.trigger_mode.get_value()) == "Internal"

    await setup_trigger_mode(DetectorTrigger.edge_trigger)
    assert (await adkinetix.drv.trigger_mode.get_value()) == "Rising Edge"

    await setup_trigger_mode(DetectorTrigger.constant_gate)
    assert (await adkinetix.drv.trigger_mode.get_value()) == "Exp. Gate"

    await setup_trigger_mode(DetectorTrigger.internal)
    assert (await adkinetix.drv.trigger_mode.get_value()) == "Internal"

    await setup_trigger_mode(DetectorTrigger.variable_gate)
    assert (await adkinetix.drv.trigger_mode.get_value()) == "Exp. Gate"


async def test_hints_from_hdf_writer(adkinetix: KinetixDetector):
    assert adkinetix.hints == {"fields": ["adkinetix"]}


async def test_can_read(adkinetix: KinetixDetector):
    # Standard detector can be used as Readable
    assert (await adkinetix.read()) == {}


async def test_decribe_describes_writer_dataset(adkinetix: KinetixDetector):
    set_mock_value(adkinetix._writer.hdf.file_path_exists, True)
    set_mock_value(adkinetix._writer.hdf.capture, True)

    assert await adkinetix.describe() == {}
    await adkinetix.stage()
    assert await adkinetix.describe() == {
        "adkinetix": {
            "source": "mock+ca://KINETIX:HDF1:FullFileName_RBV",
            "shape": (0, 0),
            "dtype": "array",
            "external": "STREAM:",
        }
    }


async def test_can_collect(
    adkinetix: KinetixDetector, static_directory_provider: DirectoryProvider
):
    directory_info = static_directory_provider()
    full_file_name = directory_info.root / directory_info.resource_dir / "foo.h5"
    set_mock_value(adkinetix.hdf.full_file_name, str(full_file_name))
    set_mock_value(adkinetix._writer.hdf.file_path_exists, True)
    set_mock_value(adkinetix._writer.hdf.capture, True)
    await adkinetix.stage()
    docs = [(name, doc) async for name, doc in adkinetix.collect_asset_docs(1)]
    assert len(docs) == 2
    assert docs[0][0] == "stream_resource"
    stream_resource = docs[0][1]
    sr_uid = stream_resource["uid"]
    assert stream_resource["data_key"] == "adkinetix"
    assert stream_resource["spec"] == "AD_HDF5_SWMR_SLICE"
    assert stream_resource["root"] == str(directory_info.root)
    assert stream_resource["resource_path"] == str(
        directory_info.resource_dir / "foo.h5"
    )
    assert stream_resource["path_semantics"] == "posix"
    assert stream_resource["resource_kwargs"] == {
        "path": "/entry/data/data",
        "multiplier": 1,
        "timestamps": "/entry/instrument/NDAttributes/NDArrayTimeStamp",
    }
    assert docs[1][0] == "stream_datum"
    stream_datum = docs[1][1]
    assert stream_datum["stream_resource"] == sr_uid
    assert stream_datum["seq_nums"] == {"start": 0, "stop": 0}
    assert stream_datum["indices"] == {"start": 0, "stop": 1}


async def test_can_decribe_collect(adkinetix: KinetixDetector):
    set_mock_value(adkinetix._writer.hdf.file_path_exists, True)
    set_mock_value(adkinetix._writer.hdf.capture, True)
    assert (await adkinetix.describe_collect()) == {}
    await adkinetix.stage()
    assert (await adkinetix.describe_collect()) == {
        "adkinetix": {
            "source": "mock+ca://KINETIX:HDF1:FullFileName_RBV",
            "shape": (0, 0),
            "dtype": "array",
            "external": "STREAM:",
        }
    }



import asyncio
import os
import subprocess
import sys
import time
from pathlib import Path
from typing import Any, Callable

import pytest
from bluesky.run_engine import RunEngine, TransitionError

from ophyd_async.core import StaticDirectoryProvider

PANDA_RECORD = str(Path(__file__).parent / "panda" / "db" / "panda.db")
INCOMPLETE_BLOCK_RECORD = str(
    Path(__file__).parent / "panda" / "db" / "incomplete_block_panda.db"
)
INCOMPLETE_RECORD = str(Path(__file__).parent / "panda" / "db" / "incomplete_panda.db")
EXTRA_BLOCKS_RECORD = str(
    Path(__file__).parent / "panda" / "db" / "extra_blocks_panda.db"
)

# Prevent pytest from catching exceptions when debugging in vscode so that break on
# exception works correctly (see: https://github.com/pytest-dev/pytest/issues/7409)
if os.getenv("PYTEST_RAISE", "0") == "1":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value


# Autouse fixture that will set all EPICS networking env vars to use lo interface
# to avoid false failures caused by things like firewalls blocking EPICS traffic.
@pytest.fixture(scope="session", autouse=True)
def configure_epics_environment():
    os.environ["EPICS_CAS_INTF_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CAS_BEACON_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CA_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CAS_AUTO_ADDR_LIST"] = "NO"
    os.environ["EPICS_CA_AUTO_BEACON_ADDR_LIST"] = "NO"

    os.environ["EPICS_PVAS_INTF_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVAS_BEACON_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVA_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVAS_AUTO_BEACON_ADDR_LIST"] = "NO"
    os.environ["EPICS_PVA_AUTO_ADDR_LIST"] = "NO"


@pytest.fixture(scope="function")
def RE(request):
    loop = asyncio.new_event_loop()
    loop.set_debug(True)
    RE = RunEngine({}, call_returns_result=True, loop=loop)

    def clean_event_loop():
        if RE.state not in ("idle", "panicked"):
            try:
                RE.halt()
            except TransitionError:
                pass
        loop.call_soon_threadsafe(loop.stop)
        RE._th.join()
        loop.close()

    request.addfinalizer(clean_event_loop)
    return RE


@pytest.fixture(scope="module", params=["pva"])
def panda_pva():
    processes = [
        subprocess.Popen(
            [
                sys.executable,
                "-m",
                "epicscorelibs.ioc",
                "-m",
                macros,
                "-d",
                PANDA_RECORD,
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        for macros in [
            "INCLUDE_EXTRA_BLOCK=,INCLUDE_EXTRA_SIGNAL=",
            "EXCLUDE_WIDTH=#,IOC_NAME=PANDAQSRVIB",
            "EXCLUDE_PCAP=#,IOC_NAME=PANDAQSRVI",
        ]
    ]
    time.sleep(2)

    for p in processes:
        assert not p.poll(), p.stdout.read()

    yield processes

    for p in processes:
        p.terminate()
        p.communicate()


@pytest.fixture
async def normal_coroutine() -> Callable[[], Any]:
    async def inner_coroutine():
        await asyncio.sleep(0.01)

    return inner_coroutine


@pytest.fixture
async def failing_coroutine() -> Callable[[], Any]:
    async def inner_coroutine():
        await asyncio.sleep(0.01)
        raise ValueError()

    return inner_coroutine


@pytest.fixture
def static_directory_provider(tmp_path: Path):
    return StaticDirectoryProvider(directory_path=tmp_path)



from __future__ import annotations

from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Protocol,
    TypeVar,
    runtime_checkable,
)

from bluesky.protocols import DataKey, HasName, Reading

if TYPE_CHECKING:
    from ophyd_async.core.async_status import AsyncStatus


@runtime_checkable
class AsyncReadable(HasName, Protocol):
    @abstractmethod
    async def read(self) -> Dict[str, Reading]:
        """Return an OrderedDict mapping string field name(s) to dictionaries
        of values and timestamps and optional per-point metadata.

        Example return value:

        .. code-block:: python

            OrderedDict(('channel1',
                         {'value': 5, 'timestamp': 1472493713.271991}),
                         ('channel2',
                         {'value': 16, 'timestamp': 1472493713.539238}))
        """
        ...

    @abstractmethod
    async def describe(self) -> Dict[str, DataKey]:
        """Return an OrderedDict with exactly the same keys as the ``read``
        method, here mapped to per-scan metadata about each field.

        Example return value:

        .. code-block:: python

            OrderedDict(('channel1',
                         {'source': 'XF23-ID:SOME_PV_NAME',
                          'dtype': 'number',
                          'shape': []}),
                        ('channel2',
                         {'source': 'XF23-ID:SOME_PV_NAME',
                          'dtype': 'number',
                          'shape': []}))
        """
        ...


@runtime_checkable
class AsyncConfigurable(Protocol):
    @abstractmethod
    async def read_configuration(self) -> Dict[str, Reading]:
        """Same API as ``read`` but for slow-changing fields related to configuration.
        e.g., exposure time. These will typically be read only once per run.
        """
        ...

    @abstractmethod
    async def describe_configuration(self) -> Dict[str, DataKey]:
        """Same API as ``describe``, but corresponding to the keys in
        ``read_configuration``.
        """
        ...


@runtime_checkable
class AsyncPausable(Protocol):
    @abstractmethod
    async def pause(self) -> None:
        """Perform device-specific work when the RunEngine pauses."""
        ...

    @abstractmethod
    async def resume(self) -> None:
        """Perform device-specific work when the RunEngine resumes after a pause."""
        ...


@runtime_checkable
class AsyncStageable(Protocol):
    @abstractmethod
    def stage(self) -> AsyncStatus:
        """An optional hook for "setting up" the device for acquisition.

        It should return a ``Status`` that is marked done when the device is
        done staging.
        """
        ...

    @abstractmethod
    def unstage(self) -> AsyncStatus:
        """A hook for "cleaning up" the device after acquisition.

        It should return a ``Status`` that is marked done when the device is finished
        unstaging.
        """
        ...


C = TypeVar("C", contravariant=True)


class Watcher(Protocol, Generic[C]):
    @staticmethod
    def __call__(
        *,
        current: C,
        initial: C,
        target: C,
        name: str | None,
        unit: str | None,
        precision: float | None,
        fraction: float | None,
        time_elapsed: float | None,
        time_remaining: float | None,
    ) -> Any: ...



import re
from dataclasses import dataclass
from inspect import isclass
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from ophyd_async.core import Device, DeviceVector, SoftSignalBackend
from ophyd_async.core.signal import Signal
from ophyd_async.core.utils import DEFAULT_TIMEOUT
from ophyd_async.epics._backend._p4p import PvaSignalBackend
from ophyd_async.epics.signal.signal import (
    epics_signal_r,
    epics_signal_rw,
    epics_signal_w,
    epics_signal_x,
)

T = TypeVar("T")
Access = FrozenSet[
    Union[Literal["r"], Literal["w"], Literal["rw"], Literal["x"], Literal["d"]]
]


def _strip_number_from_string(string: str) -> Tuple[str, Optional[int]]:
    match = re.match(r"(.*?)(\d*)$", string)
    assert match

    name = match.group(1)
    number = match.group(2) or None
    if number:
        number = int(number)
    return name, number


def _split_subscript(tp: T) -> Union[Tuple[Any, Tuple[Any]], Tuple[T, None]]:
    """Split a subscripted type into the its origin and args.

    If `tp` is not a subscripted type, then just return the type and None as args.

    """
    if get_origin(tp) is not None:
        return get_origin(tp), get_args(tp)

    return tp, None


def _strip_union(field: Union[Union[T], T]) -> Tuple[T, bool]:
    if get_origin(field) is Union:
        args = get_args(field)
        is_optional = type(None) in args
        for arg in args:
            if arg is not type(None):
                return arg, is_optional
    return field, False


def _strip_device_vector(field: Union[Type[Device]]) -> Tuple[bool, Type[Device]]:
    if get_origin(field) is DeviceVector:
        return True, get_args(field)[0]
    return False, field


@dataclass
class PVIEntry:
    """
    A dataclass to represent a single entry in the PVI table.
    This could either be a signal or a sub-table.
    """

    sub_entries: Dict[str, Union[Dict[int, "PVIEntry"], "PVIEntry"]]
    pvi_pv: Optional[str] = None
    device: Optional[Device] = None
    common_device_type: Optional[Type[Device]] = None


def _verify_common_blocks(entry: PVIEntry, common_device: Type[Device]):
    if not entry.sub_entries:
        return
    common_sub_devices = get_type_hints(common_device)
    for sub_name, sub_device in common_sub_devices.items():
        if sub_name.startswith("_") or sub_name == "parent":
            continue
        assert entry.sub_entries
        device_t, is_optional = _strip_union(sub_device)
        if sub_name not in entry.sub_entries and not is_optional:
            raise RuntimeError(
                f"sub device `{sub_name}:{type(sub_device)}` " "was not provided by pvi"
            )
        if isinstance(entry.sub_entries[sub_name], dict):
            for sub_sub_entry in entry.sub_entries[sub_name].values():  # type: ignore
                _verify_common_blocks(sub_sub_entry, sub_device)  # type: ignore
        else:
            _verify_common_blocks(
                entry.sub_entries[sub_name],
                sub_device,  # type: ignore
            )


_pvi_mapping: Dict[FrozenSet[str], Callable[..., Signal]] = {
    frozenset({"r", "w"}): lambda dtype, read_pv, write_pv: epics_signal_rw(
        dtype, "pva://" + read_pv, "pva://" + write_pv
    ),
    frozenset({"rw"}): lambda dtype, read_write_pv: epics_signal_rw(
        dtype, "pva://" + read_write_pv, write_pv="pva://" + read_write_pv
    ),
    frozenset({"r"}): lambda dtype, read_pv: epics_signal_r(dtype, "pva://" + read_pv),
    frozenset({"w"}): lambda dtype, write_pv: epics_signal_w(
        dtype, "pva://" + write_pv
    ),
    frozenset({"x"}): lambda _, write_pv: epics_signal_x("pva://" + write_pv),
}


def _parse_type(
    is_pvi_table: bool,
    number_suffix: Optional[int],
    common_device_type: Optional[Type[Device]],
):
    if common_device_type:
        # pre-defined type
        device_cls, _ = _strip_union(common_device_type)
        is_device_vector, device_cls = _strip_device_vector(device_cls)
        device_cls, device_args = _split_subscript(device_cls)
        assert issubclass(device_cls, Device)

        is_signal = issubclass(device_cls, Signal)
        signal_dtype = device_args[0] if device_args is not None else None

    elif is_pvi_table:
        # is a block, we can make it a DeviceVector if it ends in a number
        is_device_vector = number_suffix is not None
        is_signal = False
        signal_dtype = None
        device_cls = Device
    else:
        # is a signal, signals aren't stored in DeviceVectors unless
        # they're defined as such in the common_device_type
        is_device_vector = False
        is_signal = True
        signal_dtype = None
        device_cls = Signal

    return is_device_vector, is_signal, signal_dtype, device_cls


def _mock_common_blocks(device: Device, stripped_type: Optional[Type] = None):
    device_t = stripped_type or type(device)
    sub_devices = (
        (field, field_type)
        for field, field_type in get_type_hints(device_t).items()
        if not field.startswith("_") and field != "parent"
    )

    for device_name, device_cls in sub_devices:
        device_cls, _ = _strip_union(device_cls)
        is_device_vector, device_cls = _strip_device_vector(device_cls)
        device_cls, device_args = _split_subscript(device_cls)
        assert issubclass(device_cls, Device)

        is_signal = issubclass(device_cls, Signal)
        signal_dtype = device_args[0] if device_args is not None else None

        if is_device_vector:
            if is_signal:
                sub_device_1 = device_cls(SoftSignalBackend(signal_dtype))
                sub_device_2 = device_cls(SoftSignalBackend(signal_dtype))
                sub_device = DeviceVector({1: sub_device_1, 2: sub_device_2})
            else:
                sub_device = DeviceVector({1: device_cls(), 2: device_cls()})

                for sub_device_in_vector in sub_device.values():
                    _mock_common_blocks(sub_device_in_vector, stripped_type=device_cls)

            for value in sub_device.values():
                value.parent = sub_device
        else:
            if is_signal:
                sub_device = device_cls(SoftSignalBackend(signal_dtype))
            else:
                sub_device = getattr(device, device_name, device_cls())
                _mock_common_blocks(sub_device, stripped_type=device_cls)

        setattr(device, device_name, sub_device)
        sub_device.parent = device


async def _get_pvi_entries(entry: PVIEntry, timeout=DEFAULT_TIMEOUT):
    if not entry.pvi_pv or not entry.pvi_pv.endswith(":PVI"):
        raise RuntimeError("Top level entry must be a pvi table")

    pvi_table_signal_backend: PvaSignalBackend = PvaSignalBackend(
        None, entry.pvi_pv, entry.pvi_pv
    )
    await pvi_table_signal_backend.connect(
        timeout=timeout
    )  # create table signal backend

    pva_table = (await pvi_table_signal_backend.get_value())["pvi"]
    common_device_type_hints = (
        get_type_hints(entry.common_device_type) if entry.common_device_type else {}
    )

    for sub_name, pva_entries in pva_table.items():
        pvs = list(pva_entries.values())
        is_pvi_table = len(pvs) == 1 and pvs[0].endswith(":PVI")
        sub_name_split, sub_number_split = _strip_number_from_string(sub_name)
        is_device_vector, is_signal, signal_dtype, device_type = _parse_type(
            is_pvi_table,
            sub_number_split,
            common_device_type_hints.get(sub_name_split),
        )
        if is_signal:
            device = _pvi_mapping[frozenset(pva_entries.keys())](signal_dtype, *pvs)
        else:
            device = getattr(entry.device, sub_name, device_type())

        sub_entry = PVIEntry(
            device=device, common_device_type=device_type, sub_entries={}
        )

        if is_device_vector:
            # If device vector then we store sub_name -> {sub_number -> sub_entry}
            # and aggregate into `DeviceVector` in `_set_device_attributes`
            sub_number_split = 1 if sub_number_split is None else sub_number_split
            if sub_name_split not in entry.sub_entries:
                entry.sub_entries[sub_name_split] = {}
            entry.sub_entries[sub_name_split][sub_number_split] = sub_entry  # type: ignore
        else:
            entry.sub_entries[sub_name] = sub_entry

        if is_pvi_table:
            sub_entry.pvi_pv = pvs[0]
            await _get_pvi_entries(sub_entry)

    if entry.common_device_type:
        _verify_common_blocks(entry, entry.common_device_type)


def _set_device_attributes(entry: PVIEntry):
    for sub_name, sub_entry in entry.sub_entries.items():
        if isinstance(sub_entry, dict):
            sub_device = DeviceVector()  # type: ignore
            for key, device_vector_sub_entry in sub_entry.items():
                sub_device[key] = device_vector_sub_entry.device
                if device_vector_sub_entry.pvi_pv:
                    _set_device_attributes(device_vector_sub_entry)
                # Set the device vector entry to have the device vector as a parent
                device_vector_sub_entry.device.parent = sub_device  # type: ignore
        else:
            sub_device = sub_entry.device  # type: ignore
            if sub_entry.pvi_pv:
                _set_device_attributes(sub_entry)

        sub_device.parent = entry.device
        setattr(entry.device, sub_name, sub_device)


async def fill_pvi_entries(
    device: Device, root_pv: str, timeout=DEFAULT_TIMEOUT, mock=False
):
    """
    Fills a ``device`` with signals from a the ``root_pvi:PVI`` table.

    If the device names match with parent devices of ``device`` then types are used.
    """
    if mock:
        # set up mock signals for the common annotations
        _mock_common_blocks(device)
    else:
        # check the pvi table for devices and fill the device with them
        root_entry = PVIEntry(
            pvi_pv=root_pv,
            device=device,
            common_device_type=type(device),
            sub_entries={},
        )
        await _get_pvi_entries(root_entry, timeout=timeout)
        _set_device_attributes(root_entry)

    # We call set name now the parent field has been set in all of the
    # introspect-initialized devices. This will recursively set the names.
    device.set_name(device.name)


def create_children_from_annotations(
    device: Device, included_optional_fields: Tuple[str, ...] = ()
):
    """For intializing blocks at __init__ of ``device``."""
    for name, device_type in get_type_hints(type(device)).items():
        if name in ("_name", "parent"):
            continue
        device_type, is_optional = _strip_union(device_type)
        if is_optional and name not in included_optional_fields:
            continue
        is_device_vector, device_type = _strip_device_vector(device_type)
        if (
            is_device_vector
            or ((origin := get_origin(device_type)) and issubclass(origin, Signal))
            or (isclass(device_type) and issubclass(device_type, Signal))
        ):
            continue

        sub_device = device_type()
        setattr(device, name, sub_device)
        create_children_from_annotations(sub_device)



import pytest
from bluesky.run_engine import RunEngine

from ophyd_async.core import (
    DetectorTrigger,
    DeviceCollector,
    DirectoryProvider,
    TriggerInfo,
    set_mock_value,
)
from ophyd_async.epics.areadetector.drivers.pilatus_driver import PilatusTriggerMode
from ophyd_async.epics.areadetector.pilatus import PilatusDetector


@pytest.fixture
async def pilatus(
    RE: RunEngine,
    static_directory_provider: DirectoryProvider,
) -> PilatusDetector:
    async with DeviceCollector(mock=True):
        adpilatus = PilatusDetector("PILATUS:", static_directory_provider)

    return adpilatus


async def test_deadtime_invariant(
    pilatus: PilatusDetector,
):
    pilatus_controller = pilatus.controller
    # deadtime invariant with exposure time
    assert pilatus_controller.get_deadtime(0) == 2.28e-3
    assert pilatus_controller.get_deadtime(500) == 2.28e-3


@pytest.mark.parametrize(
    "detector_trigger,expected_trigger_mode",
    [
        (DetectorTrigger.internal, PilatusTriggerMode.internal),
        (DetectorTrigger.internal, PilatusTriggerMode.internal),
        (DetectorTrigger.internal, PilatusTriggerMode.internal),
    ],
)
async def test_trigger_mode_set(
    pilatus: PilatusDetector,
    detector_trigger: DetectorTrigger,
    expected_trigger_mode: PilatusTriggerMode,
):
    async def trigger_and_complete():
        await pilatus.controller.arm(num=1, trigger=detector_trigger)
        # Prevent timeouts
        set_mock_value(pilatus.controller._drv.acquire, True)

    # Default TriggerMode
    assert (await pilatus.drv.trigger_mode.get_value()) == PilatusTriggerMode.internal

    await trigger_and_complete()

    # TriggerSource changes
    assert (await pilatus.drv.trigger_mode.get_value()) == expected_trigger_mode


async def test_hints_from_hdf_writer(pilatus: PilatusDetector):
    assert pilatus.hints == {"fields": ["adpilatus"]}


async def test_unsupported_trigger_excepts(pilatus: PilatusDetector):
    with pytest.raises(
        ValueError,
        # str(EnumClass.value) handling changed in Python 3.11
        match=r"PilatusController only supports the following trigger types: .* but",
    ):
        await pilatus.prepare(TriggerInfo(1, DetectorTrigger.edge_trigger, 1, 1))



from pathlib import Path
from typing import Iterator, List

from event_model import StreamDatum, StreamResource, compose_stream_resource

from ophyd_async.core import DirectoryInfo

from ._hdfdataset import _HDFDataset


class _HDFFile:
    """
    :param directory_info: Contains information about how to construct a StreamResource
    :param full_file_name: Absolute path to the file to be written
    :param datasets: Datasets to write into the file
    """

    def __init__(
        self,
        directory_info: DirectoryInfo,
        full_file_name: Path,
        datasets: List[_HDFDataset],
    ) -> None:
        self._last_emitted = 0
        self._bundles = [
            compose_stream_resource(
                spec="AD_HDF5_SWMR_SLICE",
                root=str(directory_info.root),
                data_key=ds.name,
                resource_path=str(full_file_name.relative_to(directory_info.root)),
                resource_kwargs={
                    "path": ds.path,
                    "multiplier": ds.multiplier,
                    "timestamps": "/entry/instrument/NDAttributes/NDArrayTimeStamp",
                },
            )
            for ds in datasets
        ]

    def stream_resources(self) -> Iterator[StreamResource]:
        for bundle in self._bundles:
            yield bundle.stream_resource_doc

    def stream_data(self, indices_written: int) -> Iterator[StreamDatum]:
        # Indices are relative to resource
        if indices_written > self._last_emitted:
            indices = {
                "start": self._last_emitted,
                "stop": indices_written,
            }
            self._last_emitted = indices_written
            for bundle in self._bundles:
                yield bundle.compose_stream_datum(indices)
        return None



from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Protocol, Sequence, Union


@dataclass
class DirectoryInfo:
    """
    Information about where and how to write a file.

    The bluesky event model splits the URI for a resource into two segments to aid in
    different applications mounting filesystems at different mount points.
    The portion of this path which is relevant only for the writer is the 'root',
    while the path from an agreed upon mutual mounting is the resource_path.
    The resource_dir is used with the filename to construct the resource_path.

    :param root: Path of a root directory, relevant only for the file writer
    :param resource_dir: Directory into which files should be written, relative to root
    :param prefix: Optional filename prefix to add to all files
    :param suffix: Optional filename suffix to add to all files
    """

    root: Path
    resource_dir: Path
    prefix: Optional[str] = ""
    suffix: Optional[str] = ""


class DirectoryProvider(Protocol):
    @abstractmethod
    def __call__(self) -> DirectoryInfo:
        """Get the current directory to write files into"""


class StaticDirectoryProvider(DirectoryProvider):
    def __init__(
        self,
        directory_path: Union[str, Path],
        filename_prefix: str = "",
        filename_suffix: str = "",
        resource_dir: Optional[Path] = None,
    ) -> None:
        if resource_dir is None:
            resource_dir = Path(".")
        if isinstance(directory_path, str):
            directory_path = Path(directory_path)
        self._directory_info = DirectoryInfo(
            root=directory_path,
            resource_dir=resource_dir,
            prefix=filename_prefix,
            suffix=filename_suffix,
        )

    def __call__(self) -> DirectoryInfo:
        return self._directory_info


class NameProvider(Protocol):
    @abstractmethod
    def __call__(self) -> str:
        """Get the name to be used as a data_key in the descriptor document"""


class ShapeProvider(Protocol):
    @abstractmethod
    async def __call__(self) -> Sequence[int]:
        """Get the shape of the data collection"""



from abc import abstractmethod
from typing import Generic, Optional, Type

from bluesky.protocols import DataKey, Reading

from .utils import DEFAULT_TIMEOUT, ReadingValueCallback, T


class SignalBackend(Generic[T]):
    """A read/write/monitor backend for a Signals"""

    #: Datatype of the signal value
    datatype: Optional[Type[T]] = None

    #: Like ca://PV_PREFIX:SIGNAL
    @abstractmethod
    def source(self, name: str) -> str:
        """Return source of signal. Signals may pass a name to the backend, which can be
        used or discarded."""

    @abstractmethod
    async def connect(self, timeout: float = DEFAULT_TIMEOUT):
        """Connect to underlying hardware"""

    @abstractmethod
    async def put(self, value: Optional[T], wait=True, timeout=None):
        """Put a value to the PV, if wait then wait for completion for up to timeout"""

    @abstractmethod
    async def get_datakey(self, source: str) -> DataKey:
        """Metadata like source, dtype, shape, precision, units"""

    @abstractmethod
    async def get_reading(self) -> Reading:
        """The current value, timestamp and severity"""

    @abstractmethod
    async def get_value(self) -> T:
        """The current value"""

    @abstractmethod
    async def get_setpoint(self) -> T:
        """The point that a signal was requested to move to."""

    @abstractmethod
    def set_callback(self, callback: Optional[ReadingValueCallback[T]]) -> None:
        """Observe changes to the current value, timestamp and severity"""



from bluesky.protocols import HasHints, Hints

from ophyd_async.core import DirectoryProvider, StandardDetector
from ophyd_async.epics.areadetector.controllers.kinetix_controller import (
    KinetixController,
)
from ophyd_async.epics.areadetector.drivers import ADBaseShapeProvider
from ophyd_async.epics.areadetector.drivers.kinetix_driver import KinetixDriver
from ophyd_async.epics.areadetector.writers import HDFWriter, NDFileHDF


class KinetixDetector(StandardDetector, HasHints):
    """
    Ophyd-async implementation of an ADKinetix Detector.
    https://github.com/NSLS-II/ADKinetix
    """

    _controller: KinetixController
    _writer: HDFWriter

    def __init__(
        self,
        prefix: str,
        directory_provider: DirectoryProvider,
        drv_suffix="cam1:",
        hdf_suffix="HDF1:",
        name="",
    ):
        self.drv = KinetixDriver(prefix + drv_suffix)
        self.hdf = NDFileHDF(prefix + hdf_suffix)

        super().__init__(
            KinetixController(self.drv),
            HDFWriter(
                self.hdf,
                directory_provider,
                lambda: self.name,
                ADBaseShapeProvider(self.drv),
            ),
            config_sigs=(self.drv.acquire_time,),
            name=name,
        )

    @property
    def hints(self) -> Hints:
        return self._writer.hints



from abc import ABC, abstractmethod
from typing import Dict, Generic, Sequence, TypeVar

from bluesky.protocols import DataKey, Flyable, Preparable, Reading, Stageable

from .async_status import AsyncStatus
from .device import Device
from .signal import SignalR
from .utils import merge_gathered_dicts

T = TypeVar("T")


class TriggerLogic(ABC, Generic[T]):
    @abstractmethod
    async def prepare(self, value: T):
        """Move to the start of the flyscan"""

    @abstractmethod
    async def kickoff(self):
        """Start the flyscan"""

    @abstractmethod
    async def complete(self):
        """Block until the flyscan is done"""

    @abstractmethod
    async def stop(self):
        """Stop flying and wait everything to be stopped"""


class HardwareTriggeredFlyable(
    Device,
    Stageable,
    Preparable,
    Flyable,
    Generic[T],
):
    def __init__(
        self,
        trigger_logic: TriggerLogic[T],
        configuration_signals: Sequence[SignalR] = (),
        name: str = "",
    ):
        self._trigger_logic = trigger_logic
        self._configuration_signals = tuple(configuration_signals)
        super().__init__(name=name)

    @property
    def trigger_logic(self) -> TriggerLogic[T]:
        return self._trigger_logic

    @AsyncStatus.wrap
    async def stage(self) -> None:
        await self.unstage()

    @AsyncStatus.wrap
    async def unstage(self) -> None:
        await self._trigger_logic.stop()

    def prepare(self, value: T) -> AsyncStatus:
        """Setup trajectories"""
        return AsyncStatus(self._prepare(value))

    async def _prepare(self, value: T) -> None:
        # Move to start and setup the flyscan
        await self._trigger_logic.prepare(value)

    @AsyncStatus.wrap
    async def kickoff(self) -> None:
        await self._trigger_logic.kickoff()

    @AsyncStatus.wrap
    async def complete(self) -> None:
        await self._trigger_logic.complete()

    async def describe_configuration(self) -> Dict[str, DataKey]:
        return await merge_gathered_dicts(
            [sig.describe() for sig in self._configuration_signals]
        )

    async def read_configuration(self) -> Dict[str, Reading]:
        return await merge_gathered_dicts(
            [sig.read() for sig in self._configuration_signals]
        )



import asyncio
from unittest.mock import patch

import pytest

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    Device,
    DeviceCollector,
    SignalR,
    StaticDirectoryProvider,
    set_mock_value,
)
from ophyd_async.epics.pvi import create_children_from_annotations, fill_pvi_entries
from ophyd_async.epics.signal.signal import epics_signal_r
from ophyd_async.panda import CommonPandaBlocks
from ophyd_async.panda.writers._hdf_writer import (
    Capture,
    CaptureSignalWrapper,
    PandaHDFWriter,
    get_capture_signals,
    get_signals_marked_for_capture,
)
from ophyd_async.panda.writers._panda_hdf_file import _HDFFile


@pytest.fixture
async def panda_t():
    class CaptureBlock(Device):
        test_capture: SignalR

    class Panda(CommonPandaBlocks):
        block_a: CaptureBlock
        block_b: CaptureBlock

        def __init__(self, prefix: str, name: str = ""):
            self._prefix = prefix
            create_children_from_annotations(self)
            super().__init__(name)

        async def connect(self, mock: bool = False, timeout: float = DEFAULT_TIMEOUT):
            await fill_pvi_entries(
                self, self._prefix + "PVI", timeout=timeout, mock=mock
            )
            await super().connect(mock=mock, timeout=timeout)

    yield Panda


@pytest.fixture
async def mock_panda(panda_t):
    async with DeviceCollector(mock=True):
        mock_panda = panda_t("mock_PANDA", name="mock_panda")

    set_mock_value(
        mock_panda.block_a.test_capture,
        Capture.MinMaxMean,  # type: ignore[attr-defined]
    )

    set_mock_value(
        mock_panda.block_b.test_capture,
        Capture.No,  # type: ignore[attr-defined]
    )

    return mock_panda


@pytest.fixture
async def mock_writer(tmp_path, mock_panda) -> PandaHDFWriter:
    dir_prov = StaticDirectoryProvider(
        directory_path=str(tmp_path), filename_prefix="", filename_suffix="/data"
    )
    async with DeviceCollector(mock=True):
        writer = PandaHDFWriter(
            prefix="TEST-PANDA",
            directory_provider=dir_prov,
            name_provider=lambda: "test-panda",
            panda_device=mock_panda,
        )

    return writer


async def test_get_capture_signals_gets_all_signals(mock_panda):
    async with DeviceCollector(mock=True):
        mock_panda.test_seq = Device("seq")
        mock_panda.test_seq.seq1_capture = epics_signal_r(
            str, "pva://read_pv_1", name="seq1_capture"
        )
        mock_panda.test_seq.seq2_capture = epics_signal_r(
            str, "pva://read_pv_2", name="seq2_capture"
        )
        await asyncio.gather(
            mock_panda.test_seq.connect(mock=True),
            mock_panda.test_seq.seq1_capture.connect(mock=True),
            mock_panda.test_seq.seq2_capture.connect(mock=True),
        )
    capture_signals = get_capture_signals(mock_panda)
    expected_signals = [
        "block_a.test_capture",
        "block_b.test_capture",
        "test_seq.seq1_capture",
        "test_seq.seq2_capture",
    ]
    for signal in expected_signals:
        assert signal in capture_signals.keys()


async def test_get_signals_marked_for_capture(mock_panda):
    capture_signals = {
        "block_a.test_capture": mock_panda.block_a.test_capture,
        "block_b.test_capture": mock_panda.block_b.test_capture,
    }
    signals_marked_for_capture = await get_signals_marked_for_capture(capture_signals)
    assert len(signals_marked_for_capture) == 1
    assert signals_marked_for_capture["block_a.test"].capture_type == Capture.MinMaxMean


async def test_open_returns_correct_descriptors(mock_writer: PandaHDFWriter):
    assert hasattr(mock_writer.panda_device, "data")
    cap1 = mock_writer.panda_device.block_a.test_capture  # type: ignore[attr-defined]
    cap2 = mock_writer.panda_device.block_b.test_capture  # type: ignore[attr-defined]
    set_mock_value(cap1, Capture.MinMaxMean)
    set_mock_value(cap2, Capture.Value)
    description = await mock_writer.open()  # to make capturing status not time out
    assert len(description) == 4
    for key, entry in description.items():
        assert entry.get("shape") == [1]
        assert entry.get("dtype") == "number"
        assert isinstance(key, str)
        assert "source" in entry
        assert entry.get("external") == "STREAM:"
    expected_datakeys = [
        "test-panda-block_a-test-Min",
        "test-panda-block_a-test-Max",
        "test-panda-block_a-test-Mean",
        "test-panda-block_b-test-Value",
    ]
    for key in expected_datakeys:
        assert key in description


async def test_open_close_sets_capture(mock_writer: PandaHDFWriter):
    assert isinstance(await mock_writer.open(), dict)
    assert await mock_writer.panda_device.data.capture.get_value()
    await mock_writer.close()
    assert not await mock_writer.panda_device.data.capture.get_value()


async def test_open_sets_file_path_and_name(mock_writer: PandaHDFWriter, tmp_path):
    await mock_writer.open()
    path = await mock_writer.panda_device.data.hdf_directory.get_value()
    assert path == str(tmp_path)
    name = await mock_writer.panda_device.data.hdf_file_name.get_value()
    assert name == "mock_panda/data.h5"


async def test_open_errors_when_multiplier_not_one(mock_writer: PandaHDFWriter):
    with pytest.raises(ValueError):
        await mock_writer.open(2)


async def test_get_indices_written(mock_writer: PandaHDFWriter):
    await mock_writer.open()
    set_mock_value(mock_writer.panda_device.data.num_captured, 4)
    written = await mock_writer.get_indices_written()
    assert written == 4


async def test_wait_for_index(mock_writer: PandaHDFWriter):
    await mock_writer.open()
    set_mock_value(mock_writer.panda_device.data.num_captured, 3)
    await mock_writer.wait_for_index(3, timeout=1)
    set_mock_value(mock_writer.panda_device.data.num_captured, 2)
    with pytest.raises(TimeoutError):
        await mock_writer.wait_for_index(3, timeout=0.1)


async def test_collect_stream_docs(mock_writer: PandaHDFWriter):
    # Give the mock writer datasets
    cap1 = mock_writer.panda_device.block_a.test_capture  # type: ignore[attr-defined]
    cap2 = mock_writer.panda_device.block_b.test_capture  # type: ignore[attr-defined]
    set_mock_value(cap1, Capture.MinMaxMean)
    set_mock_value(cap2, Capture.Value)
    await mock_writer.open()

    [item async for item in mock_writer.collect_stream_docs(1)]
    assert type(mock_writer._file) is _HDFFile
    assert mock_writer._file._last_emitted == 1
    resource_doc = mock_writer._file._bundles[0].stream_resource_doc
    assert resource_doc["data_key"] == "test-panda-block_a-test-Min"
    assert "mock_panda/data.h5" in resource_doc["resource_path"]


async def test_numeric_blocks_correctly_formated(mock_writer: PandaHDFWriter):
    async def get_numeric_signal(_):
        return {
            "device.block.1": CaptureSignalWrapper(
                epics_signal_r(str, "pva://read_pv", name="Capture.Value"),
                Capture.Value,
            )
        }

    with patch(
        "ophyd_async.panda.writers._hdf_writer.get_signals_marked_for_capture",
        get_numeric_signal,
    ):
        assert "test-panda-block-1-Value" in await mock_writer.open()



from unittest.mock import patch

import pytest

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    Device,
    DeviceCollector,
    MockSignalBackend,
    NotConnected,
    SignalRW,
)
from ophyd_async.epics.signal import epics_signal_rw


class ValueErrorBackend(MockSignalBackend):
    def __init__(self, exc_text=""):
        self.exc_text = exc_text
        super().__init__(datatype=int, initial_backend=None)

    async def connect(self, timeout: float = DEFAULT_TIMEOUT):
        raise ValueError(self.exc_text)


class WorkingDummyChildDevice(Device):
    def __init__(self, name: str = "working_dummy_child_device") -> None:
        self.working_signal = SignalRW(backend=MockSignalBackend(datatype=int))
        super().__init__(name=name)


class TimeoutDummyChildDeviceCA(Device):
    def __init__(self, name: str = "timeout_dummy_child_device_ca") -> None:
        self.timeout_signal = epics_signal_rw(int, "ca://A_NON_EXISTENT_SIGNAL")
        super().__init__(name=name)


class TimeoutDummyChildDevicePVA(Device):
    def __init__(self, name: str = "timeout_dummy_child_device_pva") -> None:
        self.timeout_signal = epics_signal_rw(int, "pva://A_NON_EXISTENT_SIGNAL")
        super().__init__(name=name)


class ValueErrorDummyChildDevice(Device):
    def __init__(
        self, name: str = "value_error_dummy_child_device", exc_text=""
    ) -> None:
        self.value_error_signal = SignalRW(backend=ValueErrorBackend(exc_text=exc_text))
        super().__init__(name=name)


class DummyDeviceOneWorkingOneTimeout(Device):
    def __init__(self, name: str = "dummy_device_one_working_one_timeout") -> None:
        self.working_child_device = WorkingDummyChildDevice()
        self.timeout_child_device = TimeoutDummyChildDeviceCA()
        super().__init__(name=name)


ONE_WORKING_ONE_TIMEOUT_OUTPUT = NotConnected(
    {
        "timeout_child_device": NotConnected(
            {"timeout_signal": NotConnected("ca://A_NON_EXISTENT_SIGNAL")}
        )
    }
)


class DummyDeviceTwoWorkingTwoTimeOutTwoValueError(Device):
    def __init__(
        self,
        name: str = "dummy_device_two_working_one_timeout_two_value_error",
    ) -> None:
        self.working_child_device1 = WorkingDummyChildDevice()
        self.working_child_device2 = WorkingDummyChildDevice()
        self.timeout_child_device_ca = TimeoutDummyChildDeviceCA()
        self.timeout_child_device_pva = TimeoutDummyChildDevicePVA()
        self.value_error_child_device1 = ValueErrorDummyChildDevice(
            exc_text="Some ValueError text"
        )
        self.value_error_child_device2 = ValueErrorDummyChildDevice()
        super().__init__(name=name)


TWO_WORKING_TWO_TIMEOUT_TWO_VALUE_ERROR_OUTPUT = NotConnected(
    {
        "timeout_child_device_ca": NotConnected(
            {
                "timeout_signal": NotConnected("ca://A_NON_EXISTENT_SIGNAL"),
            }
        ),
        "timeout_child_device_pva": NotConnected(
            {"timeout_signal": NotConnected("pva://A_NON_EXISTENT_SIGNAL")}
        ),
        "value_error_child_device1": NotConnected(
            {"value_error_signal": ValueError("Some ValueError text")}
        ),
        "value_error_child_device2": NotConnected(
            {
                "value_error_signal": ValueError(),
            }
        ),
    }
)


class DummyDeviceCombiningTopLevelSignalAndSubDevice(Device):
    def __init__(
        self, name: str = "dummy_device_combining_top_level_signal_and_sub_device"
    ) -> None:
        self.timeout_signal = epics_signal_rw(int, "ca://A_NON_EXISTENT_SIGNAL")
        self.sub_device = ValueErrorDummyChildDevice(exc_text="Some ValueError text")
        super().__init__(name=name)


async def test_error_handling_connection_timeout(caplog):
    caplog.set_level(10)

    dummy_device_one_working_one_timeout = DummyDeviceOneWorkingOneTimeout()

    # This should work since the error is a connection timeout
    with pytest.raises(NotConnected) as e:
        await dummy_device_one_working_one_timeout.connect(timeout=0.01)

    assert str(e.value) == str(ONE_WORKING_ONE_TIMEOUT_OUTPUT)

    logs = caplog.get_records("call")
    assert len(logs) == 3
    assert "signal ca://A_NON_EXISTENT_SIGNAL timed out" == logs[-1].message
    assert logs[-1].levelname == "DEBUG"


async def test_error_handling_value_errors(caplog):
    """Checks that NotConnected is aggregated correctly across Devices."""

    caplog.set_level(10)

    dummy_device_two_working_one_timeout_two_value_error = (
        DummyDeviceTwoWorkingTwoTimeOutTwoValueError("dsf")
    )

    # This should fail since the error is a ValueError
    with pytest.raises(NotConnected) as e:
        (
            await dummy_device_two_working_one_timeout_two_value_error.connect(
                timeout=0.01
            ),
        )
    assert str(e.value) == str(TWO_WORKING_TWO_TIMEOUT_TWO_VALUE_ERROR_OUTPUT)

    logs = caplog.get_records("call")
    logs = [
        log
        for log in logs
        if "ophyd_async" in log.pathname and "signal" not in log.pathname
    ]
    assert len(logs) == 4

    for i in range(0, 2):
        assert (
            "device `value_error_signal` raised unexpected exception ValueError"
            == logs[i].message
        )
        assert logs[i].levelname == "ERROR"

    assert logs[2].levelname == "DEBUG"
    assert logs[3].levelname == "DEBUG"
    # These messages could come in any order
    messages = [logs[idx].message for idx in (2, 3)]
    for protocol in ("pva", "ca"):
        assert f"signal {protocol}://A_NON_EXISTENT_SIGNAL timed out" in messages


async def test_error_handling_device_collector(caplog):
    caplog.set_level(10)
    with pytest.raises(NotConnected) as e:
        # flake8: noqa
        async with DeviceCollector(timeout=0.1):
            dummy_device_two_working_one_timeout_two_value_error = (
                DummyDeviceTwoWorkingTwoTimeOutTwoValueError()
            )
            dummy_device_one_working_one_timeout = DummyDeviceOneWorkingOneTimeout()

    expected_output = NotConnected(
        {
            "dummy_device_two_working_one_timeout_two_value_error": (
                TWO_WORKING_TWO_TIMEOUT_TWO_VALUE_ERROR_OUTPUT
            ),
            "dummy_device_one_working_one_timeout": ONE_WORKING_ONE_TIMEOUT_OUTPUT,
        }
    )
    assert str(expected_output) == str(e.value)

    logs = caplog.get_records("call")
    logs = [
        log
        for log in logs
        if "ophyd_async" in log.pathname and "signal" not in log.pathname
    ]
    assert len(logs) == 5
    assert (
        logs[0].message
        == logs[1].message
        == "device `value_error_signal` raised unexpected exception ValueError"
    )
    assert logs[0].levelname == logs[1].levelname == "ERROR"

    assert logs[2].levelname == "DEBUG"
    assert logs[3].levelname == "DEBUG"
    # These messages could come in any order
    messages = [logs[idx].message for idx in (2, 3)]
    for protocol in ("pva", "ca"):
        assert f"signal {protocol}://A_NON_EXISTENT_SIGNAL timed out" in messages


def test_not_connected_error_output():
    assert str(TWO_WORKING_TWO_TIMEOUT_TWO_VALUE_ERROR_OUTPUT) == (
        "\ntimeout_child_device_ca: NotConnected:\n"
        "    timeout_signal: NotConnected: ca://A_NON_EXISTENT_SIGNAL\n"
        "timeout_child_device_pva: NotConnected:\n"
        "    timeout_signal: NotConnected: pva://A_NON_EXISTENT_SIGNAL\n"
        "value_error_child_device1: NotConnected:\n"
        "    value_error_signal: ValueError: Some ValueError text\n"
        "value_error_child_device2: NotConnected:\n"
        "    value_error_signal: ValueError\n"
    )


async def test_combining_top_level_signal_and_child_device():
    dummy_device1 = DummyDeviceCombiningTopLevelSignalAndSubDevice()
    with pytest.raises(NotConnected) as e:
        await dummy_device1.connect(timeout=0.01)
    assert str(e.value) == (
        "\ntimeout_signal: NotConnected: ca://A_NON_EXISTENT_SIGNAL\n"
        "sub_device: NotConnected:\n"
        "    value_error_signal: ValueError: Some ValueError text\n"
    )

    with pytest.raises(NotConnected) as e:
        async with DeviceCollector(timeout=0.1):
            dummy_device2 = DummyDeviceCombiningTopLevelSignalAndSubDevice()
    assert str(e.value) == (
        "\ndummy_device2: NotConnected:\n"
        "    timeout_signal: NotConnected: ca://A_NON_EXISTENT_SIGNAL\n"
        "    sub_device: NotConnected:\n"
        "        value_error_signal: ValueError: Some ValueError text\n"
    )


async def test_format_error_string_input():
    with pytest.raises(
        RuntimeError,
        match=("Unexpected type `<class 'int'>` " "expected `str` or `dict`"),
    ):
        not_connected = NotConnected(123)
        str(not_connected)

    with pytest.raises(
        RuntimeError, match=("Unexpected type `<class 'int'>`, expected an Exception")
    ):
        not_connected = NotConnected({"test": 123})
        str(not_connected)



import asyncio
from typing import Optional

from bluesky.protocols import HasHints, Hints

from ophyd_async.core import DirectoryProvider
from ophyd_async.core.async_status import AsyncStatus
from ophyd_async.core.detector import DetectorControl, DetectorTrigger, StandardDetector
from ophyd_async.epics.areadetector.drivers.ad_base import (
    ADBase,
    ADBaseShapeProvider,
    start_acquiring_driver_and_ensure_status,
)
from ophyd_async.epics.areadetector.utils import ImageMode, ad_rw, stop_busy_record
from ophyd_async.epics.areadetector.writers.hdf_writer import HDFWriter
from ophyd_async.epics.areadetector.writers.nd_file_hdf import NDFileHDF


class FooDriver(ADBase):
    def __init__(self, prefix: str, name: str = "") -> None:
        self.trigger_mode = ad_rw(str, prefix + "TriggerMode")
        super().__init__(prefix, name)


class FooController(DetectorControl):
    def __init__(self, driver: FooDriver) -> None:
        self._drv = driver

    def get_deadtime(self, exposure: float) -> float:
        # FooDetector deadtime handling
        return 0.001

    async def arm(
        self,
        num: int,
        trigger: DetectorTrigger = DetectorTrigger.internal,
        exposure: Optional[float] = None,
    ) -> AsyncStatus:
        await asyncio.gather(
            self._drv.num_images.set(num),
            self._drv.image_mode.set(ImageMode.multiple),
            self._drv.trigger_mode.set(f"FOO{trigger}"),
        )
        if exposure is not None:
            await self._drv.acquire_time.set(exposure)
        return await start_acquiring_driver_and_ensure_status(self._drv)

    async def disarm(self):
        await stop_busy_record(self._drv.acquire, False, timeout=1)


class FooDetector(StandardDetector, HasHints):
    _controller: FooController
    _writer: HDFWriter

    def __init__(
        self,
        prefix: str,
        directory_provider: DirectoryProvider,
        drv_suffix="cam1:",
        hdf_suffix="HDF1:",
        name="",
    ):
        # Must be children to pick up connect
        self.drv = FooDriver(prefix + drv_suffix)
        self.hdf = NDFileHDF(prefix + hdf_suffix)

        super().__init__(
            FooController(self.drv),
            HDFWriter(
                self.hdf,
                directory_provider,
                lambda: self.name,
                ADBaseShapeProvider(self.drv),
            ),
            config_sigs=(self.drv.acquire_time,),
            name=name,
        )

    @property
    def hints(self) -> Hints:
        return self._writer.hints



from collections import defaultdict

from bluesky import plans as bp
from bluesky.run_engine import RunEngine

from ophyd_async.core.signal import assert_emitted
from ophyd_async.sim.sim_pattern_generator import SimPatternDetector


# NOTE the async operations with h5py are non-trival
# because of lack of native support for async operations
# see https://github.com/h5py/h5py/issues/837
async def test_streaming_plan(RE: RunEngine, sim_pattern_detector: SimPatternDetector):
    names = []
    docs = []

    def append_and_print(name, doc):
        names.append(name)
        docs.append(doc)

    RE.subscribe(append_and_print)

    RE(bp.count([sim_pattern_detector], num=1))

    # NOTE - double resource because double stream
    assert names == [
        "start",
        "descriptor",
        "stream_resource",
        "stream_resource",
        "stream_datum",
        "stream_datum",
        "event",
        "stop",
    ]
    await sim_pattern_detector.writer.close()


async def test_plan(RE: RunEngine, sim_pattern_detector: SimPatternDetector):
    docs = defaultdict(list)
    RE(bp.count([sim_pattern_detector]), lambda name, doc: docs[name].append(doc))
    assert_emitted(
        docs, start=1, descriptor=1, stream_resource=2, stream_datum=2, event=1, stop=1
    )
    await sim_pattern_detector.writer.close()



from unittest.mock import patch

import pytest
from bluesky import RunEngine

from ophyd_async.core import save_device
from ophyd_async.core.device import DeviceCollector
from ophyd_async.core.utils import DEFAULT_TIMEOUT
from ophyd_async.epics.pvi import fill_pvi_entries
from ophyd_async.epics.signal import epics_signal_rw
from ophyd_async.panda import CommonPandaBlocks, TimeUnits
from ophyd_async.panda._common_blocks import DataBlock
from ophyd_async.panda._utils import phase_sorter


@pytest.fixture
async def mock_panda():
    class Panda(CommonPandaBlocks):
        data: DataBlock

        def __init__(self, prefix: str, name: str = ""):
            self._prefix = prefix
            super().__init__(name)

        async def connect(self, mock: bool = False, timeout: float = DEFAULT_TIMEOUT):
            await fill_pvi_entries(
                self, self._prefix + "PVI", timeout=timeout, mock=mock
            )
            await super().connect(mock=mock, timeout=timeout)

    async with DeviceCollector(mock=True):
        mock_panda = Panda("PANDA")
        mock_panda.phase_1_signal_units = epics_signal_rw(int, "")
    assert mock_panda.name == "mock_panda"
    yield mock_panda


@patch("ophyd_async.core.device_save_loader.save_to_yaml")
async def test_save_panda(mock_save_to_yaml, mock_panda, RE: RunEngine):
    RE(save_device(mock_panda, "path", sorter=phase_sorter))
    mock_save_to_yaml.assert_called_once()
    assert mock_save_to_yaml.call_args[0] == (
        [
            {
                "phase_1_signal_units": 0,
                "seq.1.prescale_units": TimeUnits("min"),
                "seq.2.prescale_units": TimeUnits("min"),
            },
            {
                "data.capture": False,
                "data.flush_period": 0.0,
                "data.hdf_directory": "",
                "data.hdf_file_name": "",
                "data.num_capture": 0,
                "pcap.arm": False,
                "pulse.1.delay": 0.0,
                "pulse.1.width": 0.0,
                "pulse.2.delay": 0.0,
                "pulse.2.width": 0.0,
                "seq.1.table": {},
                "seq.1.active": False,
                "seq.1.repeats": 0,
                "seq.1.prescale": 0.0,
                "seq.1.enable": "",
                "seq.2.table": {},
                "seq.2.active": False,
                "seq.2.repeats": 0,
                "seq.2.prescale": 0.0,
                "seq.2.enable": "",
            },
        ],
        "path",
    )



from ._providers import (
    DirectoryInfo,
    DirectoryProvider,
    NameProvider,
    ShapeProvider,
    StaticDirectoryProvider,
)
from .async_status import AsyncStatus, WatchableAsyncStatus
from .detector import (
    DetectorControl,
    DetectorTrigger,
    DetectorWriter,
    StandardDetector,
    TriggerInfo,
)
from .device import Device, DeviceCollector, DeviceVector
from .device_save_loader import (
    get_signal_values,
    load_device,
    load_from_yaml,
    save_device,
    save_to_yaml,
    set_signal_values,
    walk_rw_signals,
)
from .flyer import HardwareTriggeredFlyable, TriggerLogic
from .mock_signal_backend import MockSignalBackend
from .mock_signal_utils import (
    callback_on_mock_put,
    get_mock_put,
    mock_puts_blocked,
    reset_mock_put_calls,
    set_mock_put_proceeds,
    set_mock_value,
    set_mock_values,
)
from .signal import (
    Signal,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
    assert_configuration,
    assert_emitted,
    assert_reading,
    assert_value,
    observe_value,
    set_and_wait_for_value,
    soft_signal_r_and_setter,
    soft_signal_rw,
    wait_for_value,
)
from .signal_backend import SignalBackend
from .soft_signal_backend import SoftSignalBackend
from .standard_readable import ConfigSignal, HintedSignal, StandardReadable
from .utils import (
    DEFAULT_TIMEOUT,
    CalculatableTimeout,
    CalculateTimeout,
    Callback,
    NotConnected,
    ReadingValueCallback,
    T,
    get_dtype,
    get_unique,
    merge_gathered_dicts,
    wait_for_connection,
)

__all__ = [
    "get_mock_put",
    "callback_on_mock_put",
    "mock_puts_blocked",
    "set_mock_values",
    "reset_mock_put_calls",
    "SignalBackend",
    "SoftSignalBackend",
    "DetectorControl",
    "MockSignalBackend",
    "DetectorTrigger",
    "DetectorWriter",
    "StandardDetector",
    "Device",
    "DeviceCollector",
    "DeviceVector",
    "Signal",
    "SignalR",
    "SignalW",
    "SignalRW",
    "SignalX",
    "soft_signal_r_and_setter",
    "soft_signal_rw",
    "observe_value",
    "set_and_wait_for_value",
    "set_mock_put_proceeds",
    "set_mock_value",
    "wait_for_value",
    "AsyncStatus",
    "WatchableAsyncStatus",
    "DirectoryInfo",
    "DirectoryProvider",
    "NameProvider",
    "ShapeProvider",
    "StaticDirectoryProvider",
    "StandardReadable",
    "ConfigSignal",
    "HintedSignal",
    "TriggerInfo",
    "TriggerLogic",
    "HardwareTriggeredFlyable",
    "CalculateTimeout",
    "CalculatableTimeout",
    "DEFAULT_TIMEOUT",
    "Callback",
    "NotConnected",
    "ReadingValueCallback",
    "T",
    "get_dtype",
    "get_unique",
    "merge_gathered_dicts",
    "wait_for_connection",
    "get_signal_values",
    "load_from_yaml",
    "save_to_yaml",
    "set_signal_values",
    "walk_rw_signals",
    "load_device",
    "save_device",
    "assert_reading",
    "assert_value",
    "assert_configuration",
    "assert_emitted",
]



import logging
import sys
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Sequence, Type, Union

from aioca import (
    FORMAT_CTRL,
    FORMAT_RAW,
    FORMAT_TIME,
    CANothing,
    Subscription,
    caget,
    camonitor,
    caput,
)
from aioca.types import AugmentedValue, Dbr, Format
from bluesky.protocols import DataKey, Dtype, Reading
from epicscorelibs.ca import dbr

from ophyd_async.core import (
    ReadingValueCallback,
    SignalBackend,
    T,
    get_dtype,
    get_unique,
    wait_for_connection,
)
from ophyd_async.core.utils import DEFAULT_TIMEOUT, NotConnected

from .common import get_supported_values

dbr_to_dtype: Dict[Dbr, Dtype] = {
    dbr.DBR_STRING: "string",
    dbr.DBR_SHORT: "integer",
    dbr.DBR_FLOAT: "number",
    dbr.DBR_CHAR: "string",
    dbr.DBR_LONG: "integer",
    dbr.DBR_DOUBLE: "number",
}


@dataclass
class CaConverter:
    read_dbr: Optional[Dbr]
    write_dbr: Optional[Dbr]

    def write_value(self, value) -> Any:
        return value

    def value(self, value: AugmentedValue):
        return value

    def reading(self, value: AugmentedValue):
        return {
            "value": self.value(value),
            "timestamp": value.timestamp,
            "alarm_severity": -1 if value.severity > 2 else value.severity,
        }

    def get_datakey(self, source: str, value: AugmentedValue) -> DataKey:
        return {"source": source, "dtype": dbr_to_dtype[value.datatype], "shape": []}


class CaLongStrConverter(CaConverter):
    def __init__(self):
        return super().__init__(dbr.DBR_CHAR_STR, dbr.DBR_CHAR_STR)

    def write_value(self, value: str):
        # Add a null in here as this is what the commandline caput does
        # TODO: this should be in the server so check if it can be pushed to asyn
        return value + "\0"


class CaArrayConverter(CaConverter):
    def get_datakey(self, source: str, value: AugmentedValue) -> DataKey:
        return {"source": source, "dtype": "array", "shape": [len(value)]}


@dataclass
class CaEnumConverter(CaConverter):
    choices: dict[str, str]

    def write_value(self, value: Union[Enum, str]):
        if isinstance(value, Enum):
            return value.value
        else:
            return value

    def value(self, value: AugmentedValue):
        return self.choices[value]

    def get_datakey(self, source: str, value: AugmentedValue) -> DataKey:
        return {
            "source": source,
            "dtype": "string",
            "shape": [],
            "choices": list(self.choices),
        }


class DisconnectedCaConverter(CaConverter):
    def __getattribute__(self, __name: str) -> Any:
        raise NotImplementedError("No PV has been set as connect() has not been called")


def make_converter(
    datatype: Optional[Type], values: Dict[str, AugmentedValue]
) -> CaConverter:
    pv = list(values)[0]
    pv_dbr = get_unique({k: v.datatype for k, v in values.items()}, "datatypes")
    is_array = bool([v for v in values.values() if v.element_count > 1])
    if is_array and datatype is str and pv_dbr == dbr.DBR_CHAR:
        # Override waveform of chars to be treated as string
        return CaLongStrConverter()
    elif is_array and pv_dbr == dbr.DBR_STRING:
        # Waveform of strings, check we wanted this
        if datatype and datatype != Sequence[str]:
            raise TypeError(f"{pv} has type [str] not {datatype.__name__}")
        return CaArrayConverter(pv_dbr, None)
    elif is_array:
        pv_dtype = get_unique({k: v.dtype for k, v in values.items()}, "dtypes")
        # This is an array
        if datatype:
            # Check we wanted an array of this type
            dtype = get_dtype(datatype)
            if not dtype:
                raise TypeError(f"{pv} has type [{pv_dtype}] not {datatype.__name__}")
            if dtype != pv_dtype:
                raise TypeError(f"{pv} has type [{pv_dtype}] not [{dtype}]")
        return CaArrayConverter(pv_dbr, None)
    elif pv_dbr == dbr.DBR_ENUM and datatype is bool:
        # Database can't do bools, so are often representated as enums, CA can do int
        pv_choices_len = get_unique(
            {k: len(v.enums) for k, v in values.items()}, "number of choices"
        )
        if pv_choices_len != 2:
            raise TypeError(f"{pv} has {pv_choices_len} choices, can't map to bool")
        return CaConverter(dbr.DBR_SHORT, dbr.DBR_SHORT)
    elif pv_dbr == dbr.DBR_ENUM:
        # This is an Enum
        pv_choices = get_unique(
            {k: tuple(v.enums) for k, v in values.items()}, "choices"
        )
        supported_values = get_supported_values(pv, datatype, pv_choices)
        return CaEnumConverter(dbr.DBR_STRING, None, supported_values)
    else:
        value = list(values.values())[0]
        # Done the dbr check, so enough to check one of the values
        if datatype and not isinstance(value, datatype):
            raise TypeError(
                f"{pv} has type {type(value).__name__.replace('ca_', '')} "
                + f"not {datatype.__name__}"
            )
        return CaConverter(pv_dbr, None)


_tried_pyepics = False


def _use_pyepics_context_if_imported():
    global _tried_pyepics
    if not _tried_pyepics:
        ca = sys.modules.get("epics.ca", None)
        if ca:
            ca.use_initial_context()
        _tried_pyepics = True


class CaSignalBackend(SignalBackend[T]):
    def __init__(self, datatype: Optional[Type[T]], read_pv: str, write_pv: str):
        self.datatype = datatype
        self.read_pv = read_pv
        self.write_pv = write_pv
        self.initial_values: Dict[str, AugmentedValue] = {}
        self.converter: CaConverter = DisconnectedCaConverter(None, None)
        self.subscription: Optional[Subscription] = None

    def source(self, name: str):
        return f"ca://{self.read_pv}"

    async def _store_initial_value(self, pv, timeout: float = DEFAULT_TIMEOUT):
        try:
            self.initial_values[pv] = await caget(
                pv, format=FORMAT_CTRL, timeout=timeout
            )
        except CANothing as exc:
            logging.debug(f"signal ca://{pv} timed out")
            raise NotConnected(f"ca://{pv}") from exc

    async def connect(self, timeout: float = DEFAULT_TIMEOUT):
        _use_pyepics_context_if_imported()
        if self.read_pv != self.write_pv:
            # Different, need to connect both
            await wait_for_connection(
                read_pv=self._store_initial_value(self.read_pv, timeout=timeout),
                write_pv=self._store_initial_value(self.write_pv, timeout=timeout),
            )
        else:
            # The same, so only need to connect one
            await self._store_initial_value(self.read_pv, timeout=timeout)
        self.converter = make_converter(self.datatype, self.initial_values)

    async def put(self, value: Optional[T], wait=True, timeout=None):
        if value is None:
            write_value = self.initial_values[self.write_pv]
        else:
            write_value = self.converter.write_value(value)
        await caput(
            self.write_pv,
            write_value,
            datatype=self.converter.write_dbr,
            wait=wait,
            timeout=timeout,
        )

    async def _caget(self, format: Format) -> AugmentedValue:
        return await caget(
            self.read_pv,
            datatype=self.converter.read_dbr,
            format=format,
            timeout=None,
        )

    async def get_datakey(self, source: str) -> DataKey:
        value = await self._caget(FORMAT_CTRL)
        return self.converter.get_datakey(source, value)

    async def get_reading(self) -> Reading:
        value = await self._caget(FORMAT_TIME)
        return self.converter.reading(value)

    async def get_value(self) -> T:
        value = await self._caget(FORMAT_RAW)
        return self.converter.value(value)

    async def get_setpoint(self) -> T:
        value = await caget(
            self.write_pv,
            datatype=self.converter.read_dbr,
            format=FORMAT_RAW,
            timeout=None,
        )
        return self.converter.value(value)

    def set_callback(self, callback: Optional[ReadingValueCallback[T]]) -> None:
        if callback:
            assert (
                not self.subscription
            ), "Cannot set a callback when one is already set"
            self.subscription = camonitor(
                self.read_pv,
                lambda v: callback(self.converter.reading(v), self.converter.value(v)),
                datatype=self.converter.read_dbr,
                format=FORMAT_TIME,
            )
        else:
            if self.subscription:
                self.subscription.close()
            self.subscription = None



import time
from enum import Enum
from typing import Any, AsyncGenerator, AsyncIterator, Dict, Optional, Sequence
from unittest.mock import Mock

import bluesky.plan_stubs as bps
import pytest
from bluesky.protocols import DataKey, StreamAsset
from bluesky.run_engine import RunEngine
from event_model import ComposeStreamResourceBundle, compose_stream_resource

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    DetectorControl,
    DetectorTrigger,
    DetectorWriter,
    HardwareTriggeredFlyable,
    TriggerInfo,
    TriggerLogic,
)
from ophyd_async.core.detector import StandardDetector
from ophyd_async.core.signal import observe_value
from ophyd_async.epics.signal.signal import epics_signal_rw


class TriggerState(str, Enum):
    null = "null"
    preparing = "preparing"
    starting = "starting"
    stopping = "stopping"


class DummyTriggerLogic(TriggerLogic[int]):
    def __init__(self):
        self.state = TriggerState.null

    async def prepare(self, value: int):
        self.state = TriggerState.preparing
        return value

    async def kickoff(self):
        self.state = TriggerState.starting

    async def complete(self):
        self.state = TriggerState.null

    async def stop(self):
        self.state = TriggerState.stopping


class DummyWriter(DetectorWriter):
    def __init__(self, name: str, shape: Sequence[int]):
        self.dummy_signal = epics_signal_rw(int, "pva://read_pv")
        self._shape = shape
        self._name = name
        self._file: Optional[ComposeStreamResourceBundle] = None
        self._last_emitted = 0
        self.index = 0

    async def open(self, multiplier: int = 1) -> Dict[str, DataKey]:
        return {
            self._name: DataKey(
                source="soft://some-source",
                shape=self._shape,
                dtype="number",
                external="STREAM:",
            )
        }

    async def observe_indices_written(
        self, timeout=DEFAULT_TIMEOUT
    ) -> AsyncGenerator[int, None]:
        num_captured: int
        async for num_captured in observe_value(self.dummy_signal, timeout):
            yield num_captured

    async def get_indices_written(self) -> int:
        return self.index

    async def collect_stream_docs(
        self, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        if indices_written:
            if not self._file:
                self._file = compose_stream_resource(
                    spec="AD_HDF5_SWMR_SLICE",
                    root="/",
                    data_key=self._name,
                    resource_path="",
                    resource_kwargs={
                        "path": "",
                        "multiplier": 1,
                        "timestamps": "/entry/instrument/NDAttributes/NDArrayTimeStamp",
                    },
                )
                yield "stream_resource", self._file.stream_resource_doc

            if indices_written >= self._last_emitted:
                indices = {
                    "start": self._last_emitted,
                    "stop": indices_written,
                }
                self._last_emitted = indices_written
                self._last_flush = time.monotonic()
                yield "stream_datum", self._file.compose_stream_datum(indices)

    async def close(self) -> None:
        self._file = None


@pytest.fixture
async def detectors(RE: RunEngine) -> tuple[StandardDetector, StandardDetector]:
    writers = [DummyWriter("testa", (1, 1)), DummyWriter("testb", (1, 1))]
    await writers[0].dummy_signal.connect(mock=True)
    await writers[1].dummy_signal.connect(mock=True)

    async def dummy_arm_1(self=None, trigger=None, num=0, exposure=None):
        return writers[0].dummy_signal.set(1)

    async def dummy_arm_2(self=None, trigger=None, num=0, exposure=None):
        return writers[1].dummy_signal.set(1)

    detector_1: StandardDetector[Any] = StandardDetector(
        Mock(spec=DetectorControl, get_deadtime=lambda num: num, arm=dummy_arm_1),
        writers[0],
        name="detector_1",
        writer_timeout=3,
    )
    detector_2: StandardDetector[Any] = StandardDetector(
        Mock(spec=DetectorControl, get_deadtime=lambda num: num, arm=dummy_arm_2),
        writers[1],
        name="detector_2",
        writer_timeout=3,
    )

    return (detector_1, detector_2)


async def test_hardware_triggered_flyable(
    RE: RunEngine, detectors: tuple[StandardDetector]
):
    names = []
    docs = []

    def append_and_print(name, doc):
        names.append(name)
        docs.append(doc)

    RE.subscribe(append_and_print)

    trigger_logic = DummyTriggerLogic()
    flyer = HardwareTriggeredFlyable(trigger_logic, [], name="flyer")
    trigger_info = TriggerInfo(
        num=1, trigger=DetectorTrigger.constant_gate, deadtime=2, livetime=2
    )

    def flying_plan():
        yield from bps.stage_all(*detectors, flyer)
        assert flyer._trigger_logic.state == TriggerState.stopping

        # move the flyer to the correct place, before fly scanning.
        # Prepare the flyer first to get the trigger info for the detectors
        yield from bps.prepare(flyer, 1, wait=True)

        # prepare detectors second.
        for detector in detectors:
            yield from bps.prepare(
                detector,
                trigger_info,
                wait=True,
            )

        assert flyer._trigger_logic.state == TriggerState.preparing
        for detector in detectors:
            detector.controller.disarm.assert_called_once  # type: ignore

        yield from bps.open_run()
        yield from bps.declare_stream(*detectors, name="main_stream", collect=True)

        yield from bps.kickoff(flyer)
        for detector in detectors:
            yield from bps.kickoff(detector)

        yield from bps.complete(flyer, wait=False, group="complete")
        for detector in detectors:
            yield from bps.complete(detector, wait=False, group="complete")
        assert flyer._trigger_logic.state == TriggerState.null

        # Manually incremenet the index as if a frame was taken
        for detector in detectors:
            detector.writer.index += 1

        done = False
        while not done:
            try:
                yield from bps.wait(group="complete", timeout=0.5)
            except TimeoutError:
                pass
            else:
                done = True
            yield from bps.collect(
                *detectors,
                return_payload=False,
                name="main_stream",
            )
        yield from bps.wait(group="complete")
        yield from bps.close_run()

        yield from bps.unstage_all(flyer, *detectors)
        for detector in detectors:
            assert detector.controller.disarm.called  # type: ignore
        assert trigger_logic.state == TriggerState.stopping

    # fly scan
    RE(flying_plan())

    assert names == [
        "start",
        "descriptor",
        "stream_resource",
        "stream_datum",
        "stream_resource",
        "stream_datum",
        "stop",
    ]


# To do: Populate configuration signals
async def test_describe_configuration():
    flyer = HardwareTriggeredFlyable(DummyTriggerLogic(), [], name="flyer")
    assert await flyer.describe_configuration() == {}


# To do: Populate configuration signals
async def test_read_configuration():
    flyer = HardwareTriggeredFlyable(DummyTriggerLogic(), [], name="flyer")
    assert await flyer.read_configuration() == {}



import asyncio
import atexit
import logging
import time
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Sequence, Type, Union

from bluesky.protocols import DataKey, Dtype, Reading
from p4p import Value
from p4p.client.asyncio import Context, Subscription

from ophyd_async.core import (
    ReadingValueCallback,
    SignalBackend,
    T,
    get_dtype,
    get_unique,
    wait_for_connection,
)
from ophyd_async.core.utils import DEFAULT_TIMEOUT, NotConnected

from .common import get_supported_values

# https://mdavidsaver.github.io/p4p/values.html
specifier_to_dtype: Dict[str, Dtype] = {
    "?": "integer",  # bool
    "b": "integer",  # int8
    "B": "integer",  # uint8
    "h": "integer",  # int16
    "H": "integer",  # uint16
    "i": "integer",  # int32
    "I": "integer",  # uint32
    "l": "integer",  # int64
    "L": "integer",  # uint64
    "f": "number",  # float32
    "d": "number",  # float64
    "s": "string",
}


class PvaConverter:
    def write_value(self, value):
        return value

    def value(self, value):
        return value["value"]

    def reading(self, value):
        ts = value["timeStamp"]
        sv = value["alarm"]["severity"]
        return {
            "value": self.value(value),
            "timestamp": ts["secondsPastEpoch"] + ts["nanoseconds"] * 1e-9,
            "alarm_severity": -1 if sv > 2 else sv,
        }

    def get_datakey(self, source: str, value) -> DataKey:
        dtype = specifier_to_dtype[value.type().aspy("value")]
        return {"source": source, "dtype": dtype, "shape": []}

    def metadata_fields(self) -> List[str]:
        """
        Fields to request from PVA for metadata.
        """
        return ["alarm", "timeStamp"]

    def value_fields(self) -> List[str]:
        """
        Fields to request from PVA for the value.
        """
        return ["value"]


class PvaArrayConverter(PvaConverter):
    def get_datakey(self, source: str, value) -> DataKey:
        return {"source": source, "dtype": "array", "shape": [len(value["value"])]}


class PvaNDArrayConverter(PvaConverter):
    def metadata_fields(self) -> List[str]:
        return super().metadata_fields() + ["dimension"]

    def _get_dimensions(self, value) -> List[int]:
        dimensions: List[Value] = value["dimension"]
        dims = [dim.size for dim in dimensions]
        # Note: dimensions in NTNDArray are in fortran-like order
        # with first index changing fastest.
        #
        # Therefore we need to reverse the order of the dimensions
        # here to get back to a more usual C-like order with the
        # last index changing fastest.
        return dims[::-1]

    def value(self, value):
        dims = self._get_dimensions(value)
        return value["value"].reshape(dims)

    def get_datakey(self, source: str, value) -> DataKey:
        dims = self._get_dimensions(value)
        return {"source": source, "dtype": "array", "shape": dims}

    def write_value(self, value):
        # No clear use-case for writing directly to an NDArray, and some
        # complexities around flattening to 1-D - e.g. dimension-order.
        # Don't support this for now.
        raise TypeError("Writing to NDArray not supported")


@dataclass
class PvaEnumConverter(PvaConverter):
    def __init__(self, choices: dict[str, str]):
        self.choices = tuple(choices.values())

    def write_value(self, value: Union[Enum, str]):
        if isinstance(value, Enum):
            return value.value
        else:
            return value

    def value(self, value):
        return self.choices[value["value"]["index"]]

    def get_datakey(self, source: str, value) -> DataKey:
        return {
            "source": source,
            "dtype": "string",
            "shape": [],
            "choices": list(self.choices),
        }


class PvaEnumBoolConverter(PvaConverter):
    def value(self, value):
        return value["value"]["index"]

    def get_datakey(self, source: str, value) -> DataKey:
        return {"source": source, "dtype": "integer", "shape": []}


class PvaTableConverter(PvaConverter):
    def value(self, value):
        return value["value"].todict()

    def get_datakey(self, source: str, value) -> DataKey:
        # This is wrong, but defer until we know how to actually describe a table
        return {"source": source, "dtype": "object", "shape": []}  # type: ignore


class PvaDictConverter(PvaConverter):
    def reading(self, value):
        ts = time.time()
        value = value.todict()
        # Alarm severity is vacuously 0 for a table
        return {"value": value, "timestamp": ts, "alarm_severity": 0}

    def value(self, value: Value):
        return value.todict()

    def get_datakey(self, source: str, value) -> DataKey:
        raise NotImplementedError("Describing Dict signals not currently supported")

    def metadata_fields(self) -> List[str]:
        """
        Fields to request from PVA for metadata.
        """
        return []

    def value_fields(self) -> List[str]:
        """
        Fields to request from PVA for the value.
        """
        return []


class DisconnectedPvaConverter(PvaConverter):
    def __getattribute__(self, __name: str) -> Any:
        raise NotImplementedError("No PV has been set as connect() has not been called")


def make_converter(datatype: Optional[Type], values: Dict[str, Any]) -> PvaConverter:
    pv = list(values)[0]
    typeid = get_unique({k: v.getID() for k, v in values.items()}, "typeids")
    typ = get_unique(
        {k: type(v.get("value")) for k, v in values.items()}, "value types"
    )
    if "NTScalarArray" in typeid and typ == list:
        # Waveform of strings, check we wanted this
        if datatype and datatype != Sequence[str]:
            raise TypeError(f"{pv} has type [str] not {datatype.__name__}")
        return PvaArrayConverter()
    elif "NTScalarArray" in typeid or "NTNDArray" in typeid:
        pv_dtype = get_unique(
            {k: v["value"].dtype for k, v in values.items()}, "dtypes"
        )
        # This is an array
        if datatype:
            # Check we wanted an array of this type
            dtype = get_dtype(datatype)
            if not dtype:
                raise TypeError(f"{pv} has type [{pv_dtype}] not {datatype.__name__}")
            if dtype != pv_dtype:
                raise TypeError(f"{pv} has type [{pv_dtype}] not [{dtype}]")
        if "NTNDArray" in typeid:
            return PvaNDArrayConverter()
        else:
            return PvaArrayConverter()
    elif "NTEnum" in typeid and datatype is bool:
        # Wanted a bool, but database represents as an enum
        pv_choices_len = get_unique(
            {k: len(v["value"]["choices"]) for k, v in values.items()},
            "number of choices",
        )
        if pv_choices_len != 2:
            raise TypeError(f"{pv} has {pv_choices_len} choices, can't map to bool")
        return PvaEnumBoolConverter()
    elif "NTEnum" in typeid:
        # This is an Enum
        pv_choices = get_unique(
            {k: tuple(v["value"]["choices"]) for k, v in values.items()}, "choices"
        )
        return PvaEnumConverter(get_supported_values(pv, datatype, pv_choices))
    elif "NTScalar" in typeid:
        if (
            datatype
            and not issubclass(typ, datatype)
            and not (
                typ is float and datatype is int
            )  # Allow float -> int since prec can be 0
        ):
            raise TypeError(f"{pv} has type {typ.__name__} not {datatype.__name__}")
        return PvaConverter()
    elif "NTTable" in typeid:
        return PvaTableConverter()
    elif "structure" in typeid:
        return PvaDictConverter()
    else:
        raise TypeError(f"{pv}: Unsupported typeid {typeid}")


class PvaSignalBackend(SignalBackend[T]):
    _ctxt: Optional[Context] = None

    def __init__(self, datatype: Optional[Type[T]], read_pv: str, write_pv: str):
        self.datatype = datatype
        self.read_pv = read_pv
        self.write_pv = write_pv
        self.initial_values: Dict[str, Any] = {}
        self.converter: PvaConverter = DisconnectedPvaConverter()
        self.subscription: Optional[Subscription] = None

    def source(self, name: str):
        return f"pva://{self.read_pv}"

    @property
    def ctxt(self) -> Context:
        if PvaSignalBackend._ctxt is None:
            PvaSignalBackend._ctxt = Context("pva", nt=False)

            @atexit.register
            def _del_ctxt():
                # If we don't do this we get messages like this on close:
                #   Error in sys.excepthook:
                #   Original exception was:
                PvaSignalBackend._ctxt = None

        return PvaSignalBackend._ctxt

    async def _store_initial_value(self, pv, timeout: float = DEFAULT_TIMEOUT):
        try:
            self.initial_values[pv] = await asyncio.wait_for(
                self.ctxt.get(pv), timeout=timeout
            )
        except asyncio.TimeoutError as exc:
            logging.debug(f"signal pva://{pv} timed out", exc_info=True)
            raise NotConnected(f"pva://{pv}") from exc

    async def connect(self, timeout: float = DEFAULT_TIMEOUT):
        if self.read_pv != self.write_pv:
            # Different, need to connect both
            await wait_for_connection(
                read_pv=self._store_initial_value(self.read_pv, timeout=timeout),
                write_pv=self._store_initial_value(self.write_pv, timeout=timeout),
            )
        else:
            # The same, so only need to connect one
            await self._store_initial_value(self.read_pv, timeout=timeout)
        self.converter = make_converter(self.datatype, self.initial_values)

    async def put(self, value: Optional[T], wait=True, timeout=None):
        if value is None:
            write_value = self.initial_values[self.write_pv]
        else:
            write_value = self.converter.write_value(value)
        coro = self.ctxt.put(self.write_pv, {"value": write_value}, wait=wait)
        try:
            await asyncio.wait_for(coro, timeout)
        except asyncio.TimeoutError as exc:
            logging.debug(
                f"signal pva://{self.write_pv} timed out \
                          put value: {write_value}",
                exc_info=True,
            )
            raise NotConnected(f"pva://{self.write_pv}") from exc

    async def get_datakey(self, source: str) -> DataKey:
        value = await self.ctxt.get(self.read_pv)
        return self.converter.get_datakey(source, value)

    def _pva_request_string(self, fields: List[str]) -> str:
        """
        Converts a list of requested fields into a PVA request string which can be
        passed to p4p.
        """
        return f"field({','.join(fields)})"

    async def get_reading(self) -> Reading:
        request: str = self._pva_request_string(
            self.converter.value_fields() + self.converter.metadata_fields()
        )
        value = await self.ctxt.get(self.read_pv, request=request)
        return self.converter.reading(value)

    async def get_value(self) -> T:
        request: str = self._pva_request_string(self.converter.value_fields())
        value = await self.ctxt.get(self.read_pv, request=request)
        return self.converter.value(value)

    async def get_setpoint(self) -> T:
        value = await self.ctxt.get(self.write_pv, "field(value)")
        return self.converter.value(value)

    def set_callback(self, callback: Optional[ReadingValueCallback[T]]) -> None:
        if callback:
            assert (
                not self.subscription
            ), "Cannot set a callback when one is already set"

            async def async_callback(v):
                callback(self.converter.reading(v), self.converter.value(v))

            request: str = self._pva_request_string(
                self.converter.value_fields() + self.converter.metadata_fields()
            )

            self.subscription = self.ctxt.monitor(
                self.read_pv, async_callback, request=request
            )
        else:
            if self.subscription:
                self.subscription.close()
            self.subscription = None



"""EPICS Signals over CA or PVA"""

from __future__ import annotations

from enum import Enum

try:
    from .._backend._aioca import CaSignalBackend
except ImportError as ca_error:

    class CaSignalBackend:  # type: ignore
        def __init__(*args, ca_error=ca_error, **kwargs):
            raise NotImplementedError("CA support not available") from ca_error


try:
    from .._backend._p4p import PvaSignalBackend
except ImportError as pva_error:

    class PvaSignalBackend:  # type: ignore
        def __init__(*args, pva_error=pva_error, **kwargs):
            raise NotImplementedError("PVA support not available") from pva_error


class EpicsTransport(Enum):
    """The sorts of transport EPICS support"""

    #: Use Channel Access (using aioca library)
    ca = CaSignalBackend
    #: Use PVAccess (using p4p library)
    pva = PvaSignalBackend



import asyncio
from typing import Sequence

from bluesky.protocols import Triggerable

from ophyd_async.core import (
    AsyncStatus,
    ConfigSignal,
    HintedSignal,
    SignalR,
    StandardReadable,
)

from .drivers.ad_base import ADBase
from .utils import ImageMode
from .writers.nd_plugin import NDPluginBase


class SingleTriggerDet(StandardReadable, Triggerable):
    def __init__(
        self,
        drv: ADBase,
        read_uncached: Sequence[SignalR] = (),
        name="",
        **plugins: NDPluginBase,
    ) -> None:
        self.drv = drv
        self.__dict__.update(plugins)

        self.add_readables(
            [self.drv.array_counter, *read_uncached],
            wrapper=HintedSignal.uncached,
        )

        self.add_readables([self.drv.acquire_time], wrapper=ConfigSignal)

        super().__init__(name=name)

    @AsyncStatus.wrap
    async def stage(self) -> None:
        await asyncio.gather(
            self.drv.image_mode.set(ImageMode.single),
            self.drv.wait_for_plugins.set(True),
        )
        await super().stage()

    @AsyncStatus.wrap
    async def trigger(self) -> None:
        await self.drv.acquire.set(True)



from enum import Enum

from ophyd_async.epics.signal.signal import epics_signal_rw_rbv

from .ad_base import ADBase


class KinetixTriggerMode(str, Enum):
    internal = "Internal"
    edge = "Rising Edge"
    gate = "Exp. Gate"


class KinetixReadoutMode(str, Enum):
    sensitivity = 1
    speed = 2
    dynamic_range = 3


class KinetixDriver(ADBase):
    def __init__(self, prefix: str, name: str = "") -> None:
        # self.pixel_format = epics_signal_rw_rbv(PixelFormat, prefix + "PixelFormat")
        self.trigger_mode = epics_signal_rw_rbv(
            KinetixTriggerMode, prefix + "TriggerMode"
        )
        self.mode = epics_signal_rw_rbv(KinetixReadoutMode, prefix + "ReadoutPortIdx")
        super().__init__(prefix, name)



from .aravis import AravisDetector
from .kinetix import KinetixDetector
from .pilatus import PilatusDetector
from .single_trigger_det import SingleTriggerDet
from .utils import (
    FileWriteMode,
    ImageMode,
    NDAttributeDataType,
    NDAttributesXML,
)
from .vimba import VimbaDetector

__all__ = [
    "AravisDetector",
    "KinetixDetector",
    "VimbaDetector",
    "SingleTriggerDet",
    "FileWriteMode",
    "ImageMode",
    "NDAttributeDataType",
    "NDAttributesXML",
    "PilatusDetector",
]



import asyncio
from typing import Optional

from ophyd_async.core import AsyncStatus, DetectorControl, DetectorTrigger
from ophyd_async.epics.areadetector.drivers.ad_base import (
    start_acquiring_driver_and_ensure_status,
)

from ..drivers.kinetix_driver import KinetixDriver, KinetixTriggerMode
from ..utils import ImageMode, stop_busy_record

KINETIX_TRIGGER_MODE_MAP = {
    DetectorTrigger.internal: KinetixTriggerMode.internal,
    DetectorTrigger.constant_gate: KinetixTriggerMode.gate,
    DetectorTrigger.variable_gate: KinetixTriggerMode.gate,
    DetectorTrigger.edge_trigger: KinetixTriggerMode.edge,
}


class KinetixController(DetectorControl):
    def __init__(
        self,
        driver: KinetixDriver,
    ) -> None:
        self._drv = driver

    def get_deadtime(self, exposure: float) -> float:
        return 0.001

    async def arm(
        self,
        num: int,
        trigger: DetectorTrigger = DetectorTrigger.internal,
        exposure: Optional[float] = None,
    ) -> AsyncStatus:
        await asyncio.gather(
            self._drv.trigger_mode.set(KINETIX_TRIGGER_MODE_MAP[trigger]),
            self._drv.num_images.set(num),
            self._drv.image_mode.set(ImageMode.multiple),
        )
        if exposure is not None and trigger not in [
            DetectorTrigger.variable_gate,
            DetectorTrigger.constant_gate,
        ]:
            await self._drv.acquire_time.set(exposure)
        return await start_acquiring_driver_and_ensure_status(self._drv)

    async def disarm(self):
        await stop_busy_record(self._drv.acquire, False, timeout=1)



from unittest.mock import patch

import pytest

from ophyd_async.core import StaticDirectoryProvider
from ophyd_async.core.device import DeviceCollector
from ophyd_async.sim import PatternGenerator
from ophyd_async.sim.sim_pattern_detector_writer import SimPatternDetectorWriter


@pytest.fixture
async def writer(tmp_path) -> SimPatternDetectorWriter:
    async with DeviceCollector(mock=True):
        driver = PatternGenerator()
    directory = StaticDirectoryProvider(tmp_path)

    return SimPatternDetectorWriter(driver, directory)


async def test_correct_descriptor_doc_after_open(writer: SimPatternDetectorWriter):
    with patch("ophyd_async.core.signal.wait_for_value", return_value=None):
        descriptor = await writer.open()

    assert descriptor == {
        "_entry_data_data": {
            "source": "soft:///entry/data/data",
            "shape": (1, 240, 320),
            "dtype": "array",
            "external": "STREAM:",
        },
        "_entry_sum": {
            "source": "soft:///entry/sum",
            "shape": (1,),
            "dtype": "array",
            "external": "STREAM:",
        },
    }

    await writer.close()


async def test_collect_stream_docs(writer: SimPatternDetectorWriter):
    await writer.open()
    [item async for item in writer.collect_stream_docs(1)]
    assert writer.pattern_generator._handle_for_h5_file



from .signal import (
    epics_signal_r,
    epics_signal_rw,
    epics_signal_rw_rbv,
    epics_signal_w,
    epics_signal_x,
)

__all__ = [
    "epics_signal_r",
    "epics_signal_rw",
    "epics_signal_rw_rbv",
    "epics_signal_w",
    "epics_signal_x",
]



import asyncio
import time
from enum import Enum
from typing import Any, Callable, Sequence, Tuple, Type

import numpy as np
import numpy.typing as npt
import pytest
from bluesky.protocols import Reading

from ophyd_async.core import Signal, SignalBackend, T
from ophyd_async.core.soft_signal_backend import SoftSignalBackend


class MyEnum(str, Enum):
    a = "Aaa"
    b = "Bbb"
    c = "Ccc"


def integer_d(value):
    return {"dtype": "integer", "shape": []}


def number_d(value):
    return {"dtype": "number", "shape": []}


def string_d(value):
    return {"dtype": "string", "shape": []}


def enum_d(value):
    return {"dtype": "string", "shape": [], "choices": ["Aaa", "Bbb", "Ccc"]}


def waveform_d(value):
    return {"dtype": "array", "shape": [len(value)]}


class MonitorQueue:
    def __init__(self, backend: SignalBackend):
        self.backend = backend
        self.updates: asyncio.Queue[Tuple[Reading, Any]] = asyncio.Queue()
        backend.set_callback(self.add_reading_value)

    def add_reading_value(self, reading: Reading, value):
        self.updates.put_nowait((reading, value))

    async def assert_updates(self, expected_value):
        expected_reading = {
            "value": expected_value,
            "timestamp": pytest.approx(time.monotonic(), rel=0.1),
            "alarm_severity": 0,
        }
        reading, value = await self.updates.get()

        backend_value = await self.backend.get_value()
        backend_reading = await self.backend.get_reading()

        assert value == expected_value == backend_value
        assert reading == expected_reading == backend_reading

    def close(self):
        self.backend.set_callback(None)


@pytest.mark.parametrize(
    "datatype, initial_value, put_value, descriptor",
    [
        (int, 0, 43, integer_d),
        (float, 0.0, 43.5, number_d),
        (str, "", "goodbye", string_d),
        (MyEnum, MyEnum.a, MyEnum.c, enum_d),
        (npt.NDArray[np.int8], [], [-8, 3, 44], waveform_d),
        (npt.NDArray[np.uint8], [], [218], waveform_d),
        (npt.NDArray[np.int16], [], [-855], waveform_d),
        (npt.NDArray[np.uint16], [], [5666], waveform_d),
        (npt.NDArray[np.int32], [], [-2], waveform_d),
        (npt.NDArray[np.uint32], [], [1022233], waveform_d),
        (npt.NDArray[np.int64], [], [-3], waveform_d),
        (npt.NDArray[np.uint64], [], [995444], waveform_d),
        (npt.NDArray[np.float32], [], [1.0], waveform_d),
        (npt.NDArray[np.float64], [], [0.2], waveform_d),
        (Sequence[str], [], ["nine", "ten"], waveform_d),
        # Can't do long strings until https://github.com/epics-base/pva2pva/issues/17
        # (str, "longstr", ls1, ls2, string_d),
        # (str, "longstr2.VAL$", ls1, ls2, string_d),
    ],
)
async def test_soft_signal_backend_get_put_monitor(
    datatype: Type[T],
    initial_value: T,
    put_value: T,
    descriptor: Callable[[Any], dict],
):
    backend = SoftSignalBackend(datatype)

    await backend.connect()
    q = MonitorQueue(backend)
    try:
        # Check descriptor
        source = "soft://test"
        assert dict(
            source=source, **descriptor(initial_value)
        ) == await backend.get_datakey(source)
        # Check initial value
        await q.assert_updates(
            pytest.approx(initial_value) if initial_value != "" else initial_value
        )
        # Put to new value and check that
        await backend.put(put_value)
        await q.assert_updates(pytest.approx(put_value))
    finally:
        q.close()


async def test_soft_signal_backend_with_numpy_typing():
    soft_backend = SoftSignalBackend(npt.NDArray[np.float64])
    await soft_backend.connect()

    array = await soft_backend.get_value()
    assert array.shape == (0,)


async def test_soft_signal_descriptor_fails_for_invalid_class():
    class myClass:
        def __init__(self) -> None:
            pass

    soft_signal = Signal(SoftSignalBackend(myClass))
    await soft_signal.connect()

    with pytest.raises(AssertionError):
        await soft_signal._backend.get_datakey("")



import asyncio
from typing import Optional

from ophyd_async.core import (
    AsyncStatus,
    DetectorControl,
    DetectorTrigger,
    wait_for_value,
)
from ophyd_async.panda import PcapBlock


class PandaPcapController(DetectorControl):
    def __init__(self, pcap: PcapBlock) -> None:
        self.pcap = pcap

    def get_deadtime(self, exposure: float) -> float:
        return 0.000000008

    async def arm(
        self,
        num: int,
        trigger: DetectorTrigger = DetectorTrigger.constant_gate,
        exposure: Optional[float] = None,
    ) -> AsyncStatus:
        assert trigger in (
            DetectorTrigger.constant_gate,
            trigger == DetectorTrigger.variable_gate,
        ), "Only constant_gate and variable_gate triggering is supported on the PandA"
        await asyncio.gather(self.pcap.arm.set(True))
        await wait_for_value(self.pcap.active, True, timeout=1)
        return AsyncStatus(wait_for_value(self.pcap.active, False, timeout=None))

    async def disarm(self) -> AsyncStatus:
        await asyncio.gather(self.pcap.arm.set(False))
        await wait_for_value(self.pcap.active, False, timeout=1)
        return AsyncStatus(wait_for_value(self.pcap.active, False, timeout=None))
